-- common section



-- common code



--
--  Common Section
--Code common to all generated javascript
--
--
--**************Types*****************/

function isFinite(x)
	if x == null then 
		return false
	else
		return x<1E20 and x>-1E20
	end
end


CURVE_GRANULARITY = 50

Point = class()

function Point:init(x, y)
    self.x = x
    self.y = y
end




Circle = class()

function Circle:init(x,y,r)
    self.x = x
    self.y = y
    self.radius = r
end

Parabola = class()

function Parabola:init(vx,vy,fx,fy)

    self.type = "parabola";
    self.vx = vx;
    self.vy = vy;
    self.fx = fx;
    self.fy = fy;
end

LinePair = class()
function LinePair:init(A0,B0,C0,A1,B1,C1)

    self.type = "line pair";
    self.A0 = A0;
    self.B0 = B0;
    self.C0 = C0;
    self.A1 = A1;
    self.B1 = B1;
    self.C1 = C1;
end

Hyperbola = class()
function Hyperbola:init(f0x,f0y,f1x,f1y,p0x,p0y)

    self.type = "hyperbola";
    self.f0x = f0x;
    self.f0y = f0y;
    self.f1x = f1x;
    self.f1y = f1y;
    self.p0x = p0x;
    self.p0y = p0y;
end

Ellipse = class()
function Ellipse:init(centerx,centery,p0x,p0y,minor)

    self.type = "ellipse";
    self.centerx = centerx;
    self.centery = centery;
    self.p0x = p0x;
    self.p0y = p0y;
    self.minor = minor;
end



BoundingBox = class()

function BoundingBox:init(xmin, ymin, xmax, ymax)
    self.xmin = xmin
    self.ymin = ymin
    self.xmax = xmax
    self.ymax = ymax
    self.started = false
    self.iLinesStarted = false
end

function BoundingBox:extend (x, y)
    if isFinite(x) and isFinite(y) then
        if (not self.started) then
            self.xmax = x
            self.xmin = x
            self.ymax = y
            self.ymin = y
            self.started = true
        else
            if (self.xmin > x) then self.xmin = x end
            if (self.ymin > y) then self.ymin = y end
            if (self.xmax < x) then self.xmax = x end
            if (self.ymax < y) then self.ymax = y end
        end
    end
end

Tran2DMatrix = class()

function Tran2DMatrix:init(m11, m12, m21, m22, dx, dy)
    --dest and source must be of type BoundingBox
    self.m11 = m11 self.m12 = m12
    self.m21 = m21 self.m22 = m22
    self.dx = dx
    self.dy = dy
end

function Tran2DMatrix:setMatrix(source, dest)  --Note self sets matrix to invert y direction
                                               --when changing World to Screen and vice-versa
    self.m11 = (dest.xmax - dest.xmin) / (source.xmax - source.xmin)
    self.m12 = 0
    self.m21 = 0
    self.m22 = (dest.ymax - dest.ymin) / (source.ymin - source.ymax)
    self.dx = -self.m11 * source.xmin + dest.xmin
    self.dy = -self.m22 * source.ymax + dest.ymin
end


oldmathpow = math.pow

function safePow(a, b)
    if math.abs(a) < 1E-8  then
        return 0
    else
        return oldmathpow(a, b)
    end
end

math.pow = safePow

smallnumber = 0.000001

function isTiny(x)
    if math.abs(x) < smallnumber then return true else return false end
end

function clip(ob, min, max)
	for i = 1,#ob do
		if ob[i] < min then 
			ob[i] = min
		elseif ob[i] > max then
			ob[i] = max
		end
	end
end


function lerp(x0, x1, y0, y1, x)
    if x1 == x  then
        return y1
    else
        return y0 + (x - x0) * (y1 - y0) / (x1 - x0)
    end
end

function getArrowDim()
    local dim = ARROW_DIM*TEXT_SIZE_SCALE
    return Point(dim*0.8, dim*3.2)
end


function getGCD(val1, val2)
	local x,y = val1,val2
    local w
    while (y ~= 0) do
        w = x % y
        x = y
        y = w
    end
    return x
end


function parameterOnCircle(x,y,cx,cy)
    local dx = x-cx;
    local dy = y - cy;
    local t = math.atan2(dy,dx);
    return t;
end

function parameterOnFunction(x,y)

    return x;
end

function parameterOnLine(x,y,A,B,C)

    return (B*x-A*y)/(A*A+B*B);
end

function parameterOnSegment(x,y,x0,y0,x1,y1)

    local ux = x1-x0;
    local uy = y1 - y0;
    local vx = x-x0;
    local vy = y-y0;
    local len2 = ux*ux+uy*uy;
    if (isTiny(len2)) then return 0; end
    return (ux*vx+uy*vy) / len2;
end

function distanceToPointSquared(x,y,x0,y0)

    local vx = x-x0;
    local vy = y-y0;
    return vx*vx+vy*vy;
end

function distanceToPoint(x,y,x0,y0)

    local dd = distanceToPointSquared(x,y,x0,y0);
    if (isTiny(dd)) then
	return 0 
    else
	return math.sqrt(dd)
	end
end


function parameterOnCurveRecursive(x,y,f,g, start, finish, NUM_SEGS,level )

    local t = start;
    local dt = (finish-start)/ (NUM_SEGS);
    local i;
    local have_solution = false;
    local best_t;
    local best_distance;
    for i=1,NUM_SEGS+1 do
        local xx = f(t);
        local yy = g(t);
	if (isFinite(xx) and isFinite(yy)) then
		local d2 = distanceToPointSquared(x,y,xx,yy);
		if ((not have_solution) or (d2 < best_distance))  then
		   have_solution = true;
		   best_t = t;
		   best_distance = d2
	       end
	end
        t = t+dt;
    end
    if (level>0) then
        level=level-1;
        local t0 = best_t-dt;
        local t1 = best_t+dt;
        if (t0<start) then t0 = start end
        if (t1>finish) then t1 = finish end
        best_t = parameterOnCurveRecursive(x,y,f,g, t0,t1, 10,level )
    end
   return best_t;
end

function parameterOnCurve(x,y,f,g, start, finish)

    return parameterOnCurveRecursive(x,y,f,g, start, finish,100,2);
end


function transformCoord(mat, x, y)
    local p = Point( mat.m11 * x + mat.m12 * y + mat.dx, mat.m21 * x + mat.m22 * y + mat.dy)
    return p
end

function iLineExtend(wrld, A, B, C)
    if (wrld.iLinesStarted or (not wrld.started or ((wrld.xmax - wrld.xmin) == 0 or (wrld.ymax - wrld.ymin) == 0))) then
        wrld.iLinesStarted = true
        local x = (-A*C/(A*A + B*B))
        local y = (-B*C/(A*A + B*B))
        wrld:extend(x, y)
    end
end

function setMargin(wrld, margin)
    local dx = margin*(wrld.xmax - wrld.xmin)
    local dy = margin*(wrld.ymax - wrld.ymin)
    wrld.xmin = wrld.xmin - dx
    wrld.ymin = wrld.ymin - dy
    wrld.xmax = wrld.xmax + dx
    wrld.ymax = wrld.ymax + dy
end

--Correct world bounding box if it has no width or height
function correctDegenerate(wrld)
if (wrld.xmax <= wrld.xmin) then
    wrld.xmin = wrld.xmin - 0.5
    wrld.xmax = wrld.xmax + 0.5
end
if (wrld.ymax <= wrld.ymin) then
    wrld.ymin = wrld.ymin - 0.5
    wrld.ymax = wrld.ymax + 0.5
end
end

function correctAspectRatio(wrld, scrn)
    if (((scrn.xmax - scrn.xmin) ~= 0) and ((scrn.ymax - scrn.ymin) ~= 0)) then
        local dx = wrld.xmax - wrld.xmin
        local dy = wrld.ymax - wrld.ymin
        local xratio = dx / (scrn.xmax - scrn.xmin)
        local yratio = dy / (scrn.ymax - scrn.ymin)

        if (xratio > yratio) then
            local ymid = (wrld.ymin + wrld.ymax) / 2
            dy = (xratio * (scrn.ymax - scrn.ymin)) / 2
            wrld.ymin = ymid - dy
            wrld.ymax = ymid + dy
        else
            local xmid = (wrld.xmin + wrld.xmax) / 2
            dx = (yratio * (scrn.xmax - scrn.xmin)) / 2
            wrld.xmin = xmid - dx
            wrld.xmax = xmid + dx
        end
    end
end

function setMatrices(wrld, scrn, TranMat, InvTranMat)
    TranMat:setMatrix(wrld, scrn)
    InvTranMat:setMatrix(scrn, wrld)
end

function resetBounds(wrld, scrn, TranMat, InvTranMat)
    if (FIXED_BOUNDS == 0 ) or firstTimeIn then
	wrld.started = false
	setMinBounds(wrld)
	if not wrld.started then
	    wrld.xmin = 0
	    wrld.ymin = 0
	    wrld.xmax = 1
	    wrld.ymax = 1
	end
	correctDegenerate(wrld)
	if (FIXED_BOUNDS == 0) then
	    setMargin(wrld, .075)
	end
    end
    if (FIXED_ASPECT_RATIO == 1) then
        correctAspectRatio(wrld, scrn)
    end
    setMatrices(wrld, scrn, TranMat, InvTranMat)
end

-- dx dy are proportions of world extent
function panWorld(dx,dy)
    local wrld = commonGlobal.world
    local w = wrld.xmax - wrld.xmin
    local h = wrld.ymax - wrld.ymin
    local u = w*dx
    local v = h*dy
    wrld.xmax =  wrld.xmax+u
    wrld.xmin = wrld.xmin+u
    wrld.ymax = wrld.ymax+v
    wrld.ymin = wrld.ymin+v
end

function zoomCoordinate(x,x0,k)
    return x*k+x0*(1-k);
end

-- dx dy are proportions of world extent
function zoomWorld(x0,y0,scale)
    local wrld = commonGlobal.world
    wrld.xmax  = zoomCoordinate(wrld.xmax,x0,scale)
    wrld.xmin  = zoomCoordinate(wrld.xmin,x0,scale)
    wrld.ymax  = zoomCoordinate(wrld.ymax,y0,scale)
    wrld.ymin  = zoomCoordinate(wrld.ymin,y0,scale)
end




-- common drawing code


function drawPoint(mat, ctx, x, y, radius, col, drag)

    if (isFinite(x) and isFinite(y)) then
            local center = transformCoord(mat, x, y);
            if drag then
                    radius = radius * 1.5
            end
            local point = {color = col, antialiasing = false, size = 3.5, pixellist = true, type = "point" }
            point[1] = center.x
            point[2] = center.y
            ctx.objects[#ctx.objects+1] = point
    end
end

function drawLine(mat, ctx, x0, y0, x1, y1, col, thickness, arrow, linestyle)
    if (isFinite(x0) and isFinite(y0) and isFinite(x1) and isFinite(y1)) then
        local start = transformCoord(mat, x0, y0);
        local finish = transformCoord(mat, x1, y1);
        local line = { color = col, antialiasing = true, connect = true, size = thickness, type = "line", style  = linestyle}
        line[1] = start.x
        line[2] = start.y
        line[3] = finish.x
        line[4] = finish.y
        ctx.objects[#ctx.objects+1]=line
       if (arrow ~= 0) then
            drawArrow(mat, ctx, start, finish, col, thickness, arrow)
	end
   end
end

function drawArrow(mat, ctx, start, finish, col, thickness, arrow)

    local dirVec = Point(finish.x - start.x, finish.y - start.y)
    local dirMag = math.sqrt(math.pow(dirVec.x, 2) + math.pow(dirVec.y, 2))
    if dirMag < 1.0e-6 then return end
    local unitVec= Point(dirVec.x/dirMag, dirVec.y/dirMag)
    local arrowDim = getArrowDim(); --.x has half-width, .y has height
    
    ctx:submitFill(col)
    local invVec = Point(-unitVec.y, unitVec.x);
    local arrowStart, corner1, corner2, tip;
    
        if(arrow == 1 or arrow == 4) then --draw arrow directed at & touching start point
        
            arrowStart =  Point(start.x + (arrowDim.y * unitVec.x),
                              start.y + (arrowDim.y * unitVec.y))
            
            corner1 = Point(arrowStart.x + (arrowDim.x * invVec.x), 
                                arrowStart.y + (arrowDim.x * invVec.y))
            corner2 = Point(arrowStart.x - (arrowDim.x * invVec.x),
                                arrowStart.y - (arrowDim.x * invVec.y))
            
  --          ctx:submitBeginPath()
            ctx:moveTo(arrowStart.x, arrowStart.y);
            ctx:lineTo(corner1.x, corner1.y);
            ctx:lineTo(start.x, start.y);
            ctx:lineTo(corner2.x, corner2.y);
            ctx:lineTo(arrowStart.x, arrowStart.y);
        end
            
 
        if (arrow ~= 1) then --draw arrow directed at end point
            if (arrow == 3) then --draw arrow at midpoint, directed at end point
                local midpoint = Point((finish.x + start.x)/2, (finish.y + start.y)/2);
                tip = Point(midpoint.x + (arrowDim.y * unitVec.x)/2,
                                midpoint.y + (arrowDim.y * unitVec.y)/2);
            else --draw arrow at end point
                tip = finish           
	end
	    arrowStart =  Point(tip.x - (arrowDim.y * unitVec.x),
				  tip.y - (arrowDim.y * unitVec.y));
	     corner1 =  Point(arrowStart.x + (arrowDim.x * invVec.x),
				    arrowStart.y + (arrowDim.x * invVec.y));
	     corner2 =  Point(arrowStart.x - (arrowDim.x * invVec.x),
				    arrowStart.y - (arrowDim.x * invVec.y));
	    
--            ctx:submitBeginPath()
	    ctx:moveTo(arrowStart.x, arrowStart.y);
            ctx:lineTo(corner1.x, corner1.y);
            ctx:lineTo(tip.x, tip.y);
            ctx:lineTo(corner2.x, corner2.y);
            ctx:lineTo(arrowStart.x, arrowStart.y);
            
        end


            --draw arrow at midpoint, directed at end point
  
    ctx:clearFill()    
end


function drawPolyLine(mat, ctx, points, col, thickness, linestyle)
    local started = false;
    ctx.color = col
    ctx.size = thickness
    ctx.style = linestyle
    for i=1,#points do
        if (isFinite(points[i].x) and isFinite(points[i].y)) then
            local point = transformCoord(mat, points[i].x, points[i].y);
            if ( not started) then
                ctx:moveTo(point.x, point.y);
                started = true;
            else 
                ctx:lineTo(point.x, point.y);
            end
        end
    end   
end



function drawCircle(mat, ctx, x0, y0, r, filled, col, fillcol, thickness, linestyle)

	local cc = transformCoord(mat, x0, y0);
	local wpoint = Point(x0 + math.abs(r), y0);
	local point = transformCoord(mat, wpoint.x, wpoint.y)
	local dist = math.sqrt(math.pow((point.x - cc.x), 2) + math.pow((point.y - cc.y), 2));
	local circ = { fillColor = fillcol,color = col, size = thickness,  radius = dist, fill = filled,connect = true, type = "circle", style = linestyle}
	circ[1] = cc.x
	circ[2] = cc.y
	ctx.objects[#ctx.objects+1]=circ
end

function addCurveBounds(wrld,fx, fy,t0,t1)

    local NPOINTS = 50;
    local tt = t0;
    local dt = (t1 - t0)/(NPOINTS-1);
    for i = 1,NPOINTS do
        local xx = fx(tt);
        local yy = fy(tt);
        if (isFinite(xx) and isFinite(yy)) then
                wrld:extend(xx,yy);
         end
        tt = tt + dt;
    end
end

function addFunctionBounds(wrld, fy,t0,t1)

    local NPOINTS = 50;
    local tt = t0;
    local dt = (t1 - t0)/(NPOINTS-1);
    local x0 = t0;
    local y0 = fy(t0)
    for i = 1,NPOINTS do
        local xx = tt;
        local yy = fy(tt);
        if (isFinite(xx) and isFinite(yy)) then
            if (math.abs(yy-y0) / dt <2) then -- rule out really steep rises
                wrld:extend(xx,yy);
            end
        end
        tt = tt + dt;
        x0 = xx
        y0 = yy
    end
end



function drawFunctionInner(mat, ctx, fn,x0,x1,NPOINTS,allow_recursion,started)
    local scrn = commonGlobal.screenBox;
        local xx = x0;
        local dx = (x1 - x0)/(NPOINTS-1);
        local oldy = fn(xx);
 	local oldpoint = transformCoord(mat, xx,oldy);
        local i
	for i=1, NPOINTS do
            local yy = fn(xx)
		if (isFinite(yy)) then
			local point = transformCoord(mat, xx,yy);
			if ( not started) then
				ctx:moveTo(point.x, point.y);
				started = true;
			else 
                            local slope = math.abs(yy-oldy) / dx;
                            if ((slope > 50) and (allow_recursion) and ((point.y < scrn.ymax and point.y > scrn.ymin)or(oldpoint.y < scrn.ymax and oldpoint.y > scrn.ymin))) then
                               started = drawFunctionInner(mat, ctx, fn,xx-dx,xx,24,false,started) 
                            elseif (slope > 500) then  --primitive asymptote removal
                                started = false;
                            else
				ctx:lineTo(point.x, point.y);
                            end
			end
                        oldy = yy;
                        oldpoint = point;
                else
                    started = false;
                end
                xx = xx+dx;
	end
        return started;
end


function drawFunction(wrld,mat, ctx, fn,min,max, clrStr, thickness, linestyle)

    if (min == max) then
        min = wrld.xmin
        max = wrld.xmax
    else
        if (wrld.xmin > min) then min = wrld.xmin end
        if (wrld.xmax < max) then max = wrld.xmax end
    end
    ctx.size = thickness
    ctx.color = clrStr
    ctx.style = linestyle
    drawFunctionInner(mat,ctx,fn,min,max,75,true,false )
    
end

function drawCurve(mat, ctx, fx,fy,t0,t1, clrStr, thickness, linestyle)
    ctx.size = thickness
    ctx.color = clrStr
    ctx.style = linestyle
    local started = false;
    local NPOINTS = 224;
    local tt = t0;
    local dt = (t1 - t0)/(NPOINTS-1);
    for i=1,NPOINTS do
        local xx = fx(tt)
        local yy = fy(tt)
        if (isFinite(yy) and isFinite(yy)) then
            local point = transformCoord(mat, xx,yy);
            if (not started) then
                ctx:moveTo(point.x, point.y);
                started = true;
            else 
                ctx:lineTo(point.x, point.y);
    	    end
        else
            started = false;
        end
        tt = tt + dt;
    end
end



function drawString(wrld, mat, ctx, x, y, message, radius, clrStr)
    if (isFinite(x) and isFinite(y)) then
	local offset = 1.5 * radius;
	local ob = {}
	local xOffset,yOffset
	local center = {x= (wrld.xmax + wrld.xmin)/2, y =(wrld.ymax + wrld.ymin)/2 }
	if ( x <= center.x) then
		xOffset = -1 * offset;
		ob.horizontalAlignment = "right";
	else
		xOffset = offset;
		ob.horizontalAlignment = "left";
	end
	if (y >= center.y) then
		yOffset = -1 * offset;
		ob.verticalAlignment = "bottom";
	else
		yOffset = offset;
		ob.verticalAlignment = "top";
	end
	local coord = transformCoord(mat, x, y);
	ob.text = message
	ob[1] = coord.x + xOffset
	ob[2] = coord.y + yOffset
	ob.type = "string"
	ob.color = clrStr
	ctx.objects[#ctx.objects+1]=ob
    end
end

function drawInfiniteLine(wrld, mat, ctx, A, B, C, col, thickness,linestyle)

    local p0 = Point();
    local p1 = Point();
    
    if (math.abs(B) < .000001) then
        p0.x = ( -C / A );
        p1.x = ( -C / A );
        p0.y = wrld.ymin;
        p1.y = wrld.ymax;
    elseif (math.abs(A) < .000001) then
        p0.y = ( -C / B );
        p1.y = ( -C / B );
        p0.x = wrld.xmin;
        p1.x = wrld.xmax;
    else
        local xtop = ((-B * wrld.ymax) -C) / A;
        local xbottom = ((-B * wrld.ymin) -C) / A;
        local yleft = ((-A * wrld.xmin) -C) / B;
        local yright = ((-A * wrld.xmax) -C) /B;
        
        if (-B/A > 0) then
            p0.x = xbottom;
            p0.y = yleft;
            p1.x = xtop;
            p1.y = yright;
        else
            p0.x = xtop;
            p0.y = yleft;
            p1.x = xbottom;
            p1.y = yright;
        end
        if (p0.x < wrld.xmin) then
                p0.x = wrld.xmin;
        elseif (p0.x > wrld.xmax) then
                p0.x = wrld.xmax;
        end       
        if (p0.y < wrld.ymin) then
                p0.y = wrld.ymin;
        elseif (p0.y > wrld.ymax) then
                p0.y = wrld.ymax;
        end
                
        if (p1.x < wrld.xmin) then
                p1.x = wrld.xmin;
        elseif (p1.x > wrld.xmax) then
                p1.x = wrld.xmax;
        end        
        if (p1.y < wrld.ymin) then
                p1.y = wrld.ymin;
        elseif (p1.y > wrld.ymax) then
                p1.y = wrld.ymax;
        end
    end
    drawLine(mat, ctx, p0.x, p0.y, p1.x, p1.y, col, thickness, 0, linestyle);
end



-- conics

-- Conic equation: Ax^2 + By^2 + 2Cxy + 2Dx + 2Ey + F = 0
-- Conic equation: Ax^2 + By^2 + 2Cxy + 2Dx + 2Ey + F = 0
function findConic(A, B, C, D, E, F)
 -- normalize first
    local u = math.max(math.abs(A),math.abs(B),math.abs(C))
 
    if ( not isTiny(u)) then
        u = 1.0 / u;
        A = A*u;
        B = B*u;
        C = C*u;
        D = D*u;
        E = E*u;
        F = F*u;
    end
    
    -- Determinant is used to figure out the type of conic
    local detACCB = A*B - C*C; -- Used later to find type

    local detFDDA = F*A - D*D;
    local detCDBE = C*E - D*B;
    local detDAEC = D*C - A*E;
            
    local EPSILON = 1e-8;
    if (math.abs(detACCB) < EPSILON) then
        -- Parabola
        -- Find vertex and focus from ABCDEF
        
        local detBEEF = B*F - E*E;
        local detEFCD = E*D - F*C;
        local detCDBE2 = detCDBE*detCDBE;
        local detCDBE3 = detCDBE2*detCDBE;
        local detDAEC2 = detDAEC*detDAEC;
        local detDAEC3 = detDAEC2*detDAEC;
        
        local fdenom = detCDBE2 + detDAEC2;
        local vdenom = 2*fdenom*fdenom;
        
        local vx = (detCDBE3*detBEEF + 2*detDAEC3*detEFCD + detDAEC2*detCDBE*(2*detBEEF - detFDDA)) / vdenom;
        local vy = (detDAEC3*detFDDA + 2*detCDBE3*detEFCD + detCDBE2*detDAEC*(2*detFDDA - detBEEF)) / vdenom;
        
        local fx = (detEFCD*detDAEC + 0.5*detCDBE*(detBEEF - detFDDA)) / fdenom;
        local fy = (detEFCD*detCDBE + 0.5*detDAEC*(detFDDA - detBEEF)) / fdenom;
                        
        return Parabola(vx, vy, fx, fy);
        
    end
    
    -- Find center and two foci, used for drawing
    local center =  Point(detCDBE / detACCB, detDAEC / detACCB);
  
    local cr = A*A + B*B + 4*C*C - 2*A*B;
    cr = math.sqrt(cr);
    
    local u;
    local uu;
    local vv;
    if (C == 0) then
        u =  Point(1, 0);
        uu = A;
        vv = B;
    else
        u =  Point(0.5*(A-B+cr)/C,1);
        local ul = math.sqrt(u.x*u.x + u.y*u.y);
        u =  Point(u.x / ul, u.y / ul);
        uu = 0.5*(A+B+cr);
        vv = 0.5*(A+B-cr);
    end
    
    if (uu == 0 or vv == 0) then
        return;
    end
    
    local ff = F + D*center.x + E*center.y;
    if (math.abs(ff) < EPSILON) then
        -- Line pair
         
        local B0 = C + math.sqrt(-detACCB);
        local C0 = D - math.sqrt(-detFDDA);
        local B1 = C - math.sqrt(-detACCB);
        local C1 = D + math.sqrt(-detFDDA);
        
        uu = B0*C1 + B1*C0;
        vv = 2*A*E;
        if (not (math.abs(uu-vv) < EPSILON)) then
            -- "Chosen the wrong root"
            B0 = C + math.sqrt(-detACCB);
            C0 = D + math.sqrt(-detFDDA);
            B1 = C - math.sqrt(-detACCB);
            C1 = D - math.sqrt(-detFDDA);
        end
        return LinePair(A,B0,C0,A,B1,C1)
    end
    
    local fuu = -ff/uu;
    local fvv = -ff/vv;
    uu = math.sqrt(math.abs(fuu));
    vv = math.sqrt(math.abs(fvv));
    
    local v =  Point(-u.y, u.x); -- u-perp
      
    local p0;
    local p1;
    -- Extract the two foci!
    if (fuu < 0) then -- Hyperbola needs foci swapped
        p0 =  Point(center.x + vv * v.x, center.y + vv * v.y);
        p1 =  Point(center.x + uu * u.x, center.y + uu * u.y);
    elseif (fvv > fuu) then
        p0 =  Point(center.x + vv * v.x, center.y + vv * v.y);
        p1 =  Point(center.x - uu * u.x, center.y - uu * u.y);
    else
        --assert((u^v) >= 0);
        p0 =  Point(center.x + uu * u.x, center.y + uu * u.y);
        p1 =  Point(center.x + vv * v.x, center.y + vv * v.y);
    end
    
    -- Find foci (see coregeom.cpp -> s_ellipse::get_foci)
    local cp0 =  Point(p0.x-center.x, p0.y-center.y);
    local cp1 =  Point(p1.x-center.x, p1.y-center.y);
    local d0 = math.sqrt(cp0.x*cp0.x + cp0.y*cp0.y);
    local d1 = math.sqrt(cp1.x*cp1.x + cp1.y*cp1.y);
    
    -- axis unit vector
    local cp0u =  Point(cp0.x / d0, cp0.y / d0);
    
    if (detACCB < 0) then
        -- Hyperbola
        
        local alpha = math.sqrt(d0*d0+d1*d1);
        local f0 =  Point(center.x - alpha * cp0u.x, center.y - alpha * cp0u.y);
        local f1 =  Point(center.x + alpha * cp0u.x, center.y + alpha * cp0u.y);
        
        return Hyperbola(f0.x, f0.y, f1.x, f1.y, p0.x, p0.y);
    else
        -- Ellipse
        
        local minor = math.sqrt(cp1.x*cp1.x + cp1.y*cp1.y);
        return Ellipse(center.x, center.y, p0.x, p0.y, minor);
    end
end





-- Conic equation: Ax^2 + By^2 + 2Cxy + 2Dx + 2Ey + F = 0
function drawConic(wrld, mat, ctx, A, B, C, D, E, F, clrStr, thickness, linestyle)

    local conic = findConic( A, B, C, D, E, F)
    if (conic.type == "parabola") then
	drawParabola(wrld, mat, ctx, conic.vx, conic.vy, conic.fx, conic.fy, clrStr, thickness, linestyle);      
    elseif (conic.type == "line pair") then
        drawInfiniteLine(wrld, mat, ctx, conic.A0, conic.B0, conic.C0, clrStr, thickness, linestyle);
        drawInfiniteLine(wrld, mat, ctx, conic.A1, conic.B1, conic.C1, clrStr, thickness, linestyle);       
    elseif (conic.type == "hyperbola") then
	drawHyperbola(wrld, mat, ctx, conic.f0x, conic.f0y, conic.f1x, conic.f1y, conic.p0x, conic.p0y, clrStr, thickness, linestyle);
    elseif (conic.type == "ellipse") then
	drawEllipse(wrld, mat, ctx, conic.centerx, conic.centery, conic.p0x, conic.p0y, conic.minor, clrStr, thickness, linestyle);
    end
end

-- Draw an ellipse at center (cx,cy); major point on ellipse (mx,my), minor
function drawEllipse(wrld, mat, ctx, cx, cy, mx, my, minor, col, thickness, linestyle)
    ctx.color = col
    ctx.size = thickness
    ctx.style = linestyle
    local major = math.sqrt((mx-cx)^2 + (my-cy)^2)

    local center = transformCoord(mat, cx,cy);
    local p = transformCoord(mat, mx,my);
    
    ctx:moveTo(p.x,p.y)
    local a = math.sqrt((p.x - center.x)^2 + (p.y-center.y)^2)
    local b = minor*a/major
    local ux,uy = (p.x - center.x) / a , (p.y-center.y) / a
    local vx, vy = -uy, ux
    for i = 1,100 do
        
        local xx = a*math.cos(i*math.pi/50)
        local yy = b*math.sin(i*math.pi/50)
        local x,y = center.x + xx*ux + yy*vx, center.y + xx*uy+yy*vy 
        
        ctx:lineTo(x,y);
    end
end


-- Find the parameter of parabola with vertex V, focus F where it intersects
-- with line L = Q + sD
function intersectParabola(V, F, Q, D)
    local result =  {};

    local VF =  Point(F.x - V.x, F.y - V.y);
    local R = math.sqrt(VF.x*VF.x + VF.y*VF.y);

    -- Axis
    local A =  Point(VF.x / R, VF.y / R);
    -- D perpendicular
    local Dp =  Point(-D.y, D.x);

    -- Quadratic equation
    local ADotD = A.x*D.x + A.y*D.y;
    local ADotDp = A.x*Dp.x + A.y*Dp.y;
    local QV =  Point(V.x - Q.x, V.y - Q.y);
    local DpDotQV = Dp.x*QV.x + Dp.y*QV.y;
    local oneOverR = 1 / R;
    local oneOver2R = 0.5 * oneOverR;

    local discrim = ADotD*ADotD - oneOverR * ADotDp * DpDotQV;
    if (discrim < 0) then
            return result; -- No intersections
    end       
    local denom = oneOver2R * ADotDp;
    if (denom == 0) then
            return result; -- No intersections
    end        
    local sqrtDiscrim = math.sqrt(discrim);

    result[1] = (-ADotD - sqrtDiscrim) / denom;
    result[2] = (-ADotD + sqrtDiscrim) / denom;

    return result;
end

function evaluateParabola(V, F, t)
    -- Parabola "radius" is the distance between focus and vertex
    local oldAxis =  Point(F.x - V.x, F.y - V.y);
    local radius = math.sqrt(oldAxis.x * oldAxis.x + oldAxis.y * oldAxis.y);
    local axis =  Point(oldAxis.x / radius, oldAxis.y / radius);
    local perpAxis =  Point(-axis.y, axis.x);
     local x = t;
    local denom = 0.25 / radius; -- 1 / (4 * radius)
    local y = t * t * denom;
    local a =  Point(x,y);   
    -- Control points, rotated
    -- pWorld = vertex + p.x * perpAxis + p.y * axis
    local aWorld =  Point(V.x + a.x * perpAxis.x + a.y * axis.x, V.y + a.x * perpAxis.y + a.y * axis.y);
    return aWorld
end


-- Draw section (left->right) of parabola with vertex V and focus F
function drawParabolaSection(wrld, mat, ctx, V, F, left, right, clrStr, thickness, linestyle)
    -- Parabola "radius" is the distance between focus and vertex
    local oldAxis =  Point(F.x - V.x, F.y - V.y);
    local radius = math.sqrt(oldAxis.x * oldAxis.x + oldAxis.y * oldAxis.y);
    local axis =  Point(oldAxis.x / radius, oldAxis.y / radius);
    local perpAxis =  Point(-axis.y, axis.x);
    
    local x0 = left;
    local x1 = 0.5 * (left + right);
    local x2 = right;
    
    local denom = 0.25 / radius; -- 1 / (4 * radius)
    local y0 = left * left * denom;
    local y1 = left * right * denom;
    local y2 = right * right * denom;
    
    local a =  Point(x0, y0);
    local b =  Point(x1, y1);
    local c =  Point(x2, y2);
    
    -- Control points, rotated
    -- pWorld = vertex + p.x * perpAxis + p.y * axis
    local aWorld =  Point(V.x + a.x * perpAxis.x + a.y * axis.x, V.y + a.x * perpAxis.y + a.y * axis.y);
    local bWorld =  Point(V.x + b.x * perpAxis.x + b.y * axis.x, V.y + b.x * perpAxis.y + b.y * axis.y);
    local cWorld =  Point(V.x + c.x * perpAxis.x + c.y * axis.x, V.y + c.x * perpAxis.y + c.y * axis.y);

    local aScreen = transformCoord(mat, aWorld.x, aWorld.y);
    local bScreen = transformCoord(mat, bWorld.x, bWorld.y);
    local cScreen = transformCoord(mat, cWorld.x, cWorld.y);

    -- Draw the parabola as a quadratic bezier curve
    ctx:moveTo(aScreen.x, aScreen.y);
    ctx:quadraticCurveTo(bScreen.x, bScreen.y, cScreen.x, cScreen.y);
end

-- Draw parabola with vertex (vx,vy) and focus (fx,fy)
function drawParabola(wrld, mat, ctx, vx, vy, fx, fy, col, thickness, linestyle) 
local EPSILON = 0.000001

    ctx.color = col
    ctx.size = thickness
	ctx.style = linestyle

    local vertex =  Point(vx, vy);
    local focus =  Point(fx, fy);
    
    -- Parabola "radius" is the distance between focus and vertex
    local oldAxis =  Point(fx - vx, fy - vy);
    local radius = math.sqrt(oldAxis.x * oldAxis.x + oldAxis.y * oldAxis.y);
    local axis =  Point(oldAxis.x / radius, oldAxis.y / radius);
    local perpAxis =  Point(-axis.y, axis.x);

    -- Find min and max intersect for left and right
    local intersects =  {};

    local V =  Point(vx, vy);
    local F =  Point(fx, fy);

    -- Left
    local Q =  Point(wrld.xmin, wrld.ymin);
    local dir =  Point(0, 1);
    local newIntersects = intersectParabola(V, F, Q, dir);
    for i=1,#newIntersects do
        local tt = newIntersects[i]
        local uu = evaluateParabola(vertex,focus,tt)
        if uu.y > wrld.ymin-EPSILON and uu.y < wrld.ymax + EPSILON then
            intersects[#intersects+1] = newIntersects[i];
        end
    end
    -- Top
    Q =  Point(wrld.xmin, wrld.ymax);
    dir =  Point(1, 0);
    newIntersects = intersectParabola(V, F, Q, dir);
    for i=1,#newIntersects do
        local tt = newIntersects[i]
        local uu = evaluateParabola(vertex,focus,tt)
        if uu.x > wrld.xmin-EPSILON and uu.x < wrld.xmax + EPSILON then
            intersects[#intersects+1] = newIntersects[i];
        end
    end
    -- Right
    Q =  Point(wrld.xmax, wrld.ymin);
    dir =  Point(0, 1);
    newIntersects = intersectParabola(V, F, Q, dir);
    for i=1,#newIntersects do
        local tt = newIntersects[i]
        local uu = evaluateParabola(vertex,focus,tt)
        if uu.y > wrld.ymin-EPSILON and uu.y < wrld.ymax + EPSILON then
            intersects[#intersects+1] = newIntersects[i];
        end
    end
    
    -- Bottom
    Q =  Point(wrld.xmin, wrld.ymin);
    dir =  Point(1, 0);
    newIntersects = intersectParabola(V, F, Q, dir);
    for i=1,#newIntersects do
        local tt = newIntersects[i]
        local uu = evaluateParabola(vertex,focus,tt)
        if uu.x > wrld.xmin-EPSILON and uu.x < wrld.xmax + EPSILON then
            intersects[#intersects+1] = newIntersects[i];
        end
    end
    -- Sort intersects from leftmost to rightmost
    table.sort( intersects, function(a,b) return a>b end );
    
    -- Draw all sections between the intersections; this should at least allow
    -- high-quality drawing of parabola sections in the visible regions
    -- TODO: Ideally we would only try to draw the visible sections
    for i=1,#intersects-1 do
            drawParabolaSection(wrld, mat, ctx, vertex, focus, intersects[i], intersects[i+1], clrStr, thickness);
    end
end

-- Draw hyperbola with focus 0 (f0x,f0y), focus 1 (f1x,f1y) and point (px,py)
function drawHyperbola(wrld, mat, ctx, f0x, f0y, f1x, f1y, px, py, clrStr, thickness, linestyle)
    -- Gather parameters
    local f0 =  Point(f0x, f0y);
    local f1 =  Point(f1x, f1y);
    local p =  Point(px, py);
    
    local f0f1 =  Point(f1.x-f0.x, f1.y-f0.y);
    local f0f1Len = math.sqrt(f0f1.x*f0f1.x + f0f1.y*f0f1.y);
    local f0f1Unit =  Point(f0f1.x / f0f1Len, f0f1.y / f0f1Len);
    local f0f1UnitPerp =  Point(-f0f1Unit.y, f0f1Unit.x);
    
    local f0p =  Point(p.x-f0.x, p.y-f0.y);
    local f1p =  Point(p.x-f1.x, p.y-f1.y);
    local d0 = math.sqrt(f0p.x*f0p.x + f0p.y*f0p.y);
    local d1 = math.sqrt(f1p.x*f1p.x + f1p.y*f1p.y);
    
    -- major?
    local a = 0.5 * math.abs(d0 - d1);
    -- minor?
    local f = 0.5 * f0f1Len;
    local b = math.sqrt(f*f-a*a);
    
    -- Get "virtual origin" which is halfway between the foci
    local ctr =  Point(0.5*(f0.x+f1.x), 0.5*(f0.y+f1.y));
    
    -- Array of points that approximate the hyperbola
    local arm1 =  {};
    local arm2 =  {};
    
    -- Number of points to approximate with...
    local NUM_STEPS = 128;
    
    -- Generate points
    -- TODO: Based on world bounds; only draw the parts that are visible
    local EPSILON = 1e-2;
    local umin = -0.5*math.pi+EPSILON;
    local umax = 0.5*math.pi-EPSILON;
    
    for i = 1,NUM_STEPS do
        local u = lerp(1, NUM_STEPS, umin, umax, i);
        
        local origX = a/math.cos(u);
        local origY = b*math.tan(u);
        
        -- Rotate the "x axis" along f0f1
        local x = ctr.x + origX * f0f1Unit.x + origY * f0f1UnitPerp.x;
        local y = ctr.y + origX * f0f1Unit.y + origY * f0f1UnitPerp.y;
        arm1[i] =  Point(x, y);
        
        local x = ctr.x - origX * f0f1Unit.x - origY * f0f1UnitPerp.x;
        local y = ctr.y - origX * f0f1Unit.y - origY * f0f1UnitPerp.y;
        arm2[i] =  Point(x, y);
    end	
    -- Draw all points as a polyline
    drawPolyLine(mat, ctx, arm1, clrStr, thickness, linestyle);
    drawPolyLine(mat, ctx, arm2, clrStr, thickness, linestyle);
end

-- Conic equation: Ax^2 + By^2 + 2Cxy + 2Dx + 2Ey + F = 0
function parameterOnConic(x,y, A, B, C, D, E, F,ff,gg)
    local conic = findConic( A, B, C, D, E, F)
    if (conic.type == "parabola") then
	return parameterOnParabola(x,y,conic.vx, conic.vy, conic.fx, conic.fy);      
    elseif (conic.type == "hyperbola") then
         return parameterOnHyperbola(x,y,ff,gg);
    else
	return parameterOnCurve(x,y,ff,gg,0,2*math.pi);
    end
end

function parameterOnParabola(x,y,vx,vy,fx,fy)
	local u = fx-vx;
	local v = fy-vy;
	local a = x-vx;
	local b = y-vy;
	local dd = u*u+v*v;
	if (isTiny(dd)) then
            return 0
        else
            return (a*v-b*u)/(2*dd);
        end
end

function parameterOnHyperbola(x,y,ff,gg)
    local s0 = 1.57;  -- deliberately a bit under pi/2 to avoid degenerate situation
    local t0 = parameterOnCurve(x,y,ff,gg,-s0,s0);
    if (t0 <0) then
        t0 = t0 + 2*math.pi;
    end
    local t1 = parameterOnCurve(x,y,ff,gg,math.pi-s0,math.pi+s0);
    local x0 = ff(t0);
    local y0 = gg(t0);
    local d0 = distanceToPointSquared(x,y,x0,y0);
    local x1 = ff(t1);
    local y1 = gg(t1);
    local d1 = distanceToPointSquared(x,y,x1,y1);
    if (d0<d1) then return t0; else return t1; end
end




function drawAxis(wrld, mat, invMat, ctx, A, B, C, clrStr, thickness, type, units, labels, clrTxt)
    drawInfiniteLine(wrld, mat, ctx, A, B, C, clrStr, thickness);
    local label = {};
    label.symbol = "";
    label.divisor = 1;
    label.magnitude = 1;
    local dist = getGridSpacing(wrld, units, label,type);
    local ticHalf = 5;
    
    if type == 0 then --X-axis
        if (wrld.ymin < 0) and (wrld.ymax > 0) then
            local posLo = math.ceil(wrld.xmin / dist);
            local posHi = math.ceil(wrld.xmax / dist);
            for i = posLo,posHi do
                local pos = i * dist;
                drawAxisTic(mat, ctx, type, pos, clrStr, thickness, ticHalf);
                if (labels) then
                     drawTicLabel(mat, ctx, type, pos, dist, units, label, clrTxt, ticHalf);
                end
            end
        end
    else
        if (wrld.xmin < 0) and (wrld.xmax > 0) then
            local posLo = math.ceil(wrld.ymin / dist);
            local posHi = math.ceil(wrld.ymax / dist);
            for i = posLo,posHi do
                local pos = i * dist;
                drawAxisTic(mat, ctx, type, pos, clrStr, thickness, ticHalf);
                if labels then
                    drawTicLabel(mat, ctx, type, pos, dist, units, label, clrTxt, ticHalf);
                end
            end
        end
    end
end


function getGridSpacing(wrld, units, label,axis)
    local gspacing = 0;
    local width = wrld.xmax - wrld.xmin;
    local height = wrld.ymax - wrld.ymin;
    local size = (axis == 0) and width or height;
    local degToRad = (2.0*math.pi)/360.0;

    -- see what the ideal spacing would be if we wanted AXIS_IDEAL_LABEL_COUNT on the screen
    local idealspacing = (size / commonGlobal.axisIdealLabelCount);
    
    -- idealspacing needs to be in the range of 1 to 10 in order for spacing to work
	-- save the magnitude and apply it to gspacing at the end
    local magnitude = 1.0;
    
    if (units == 0) then									-- decimal spacing
	local power = math.floor(math.log(math.abs(idealspacing)) / math.log(10));
	magnitude = math.pow(10.0, power);
        label.magnitude = magnitude;
	idealspacing = idealspacing / magnitude;

        -- tics will be some magnitude of 10, 5, 2 or 1
        -- see which division is closest to the ideal spacing
        local tics10 = math.abs(idealspacing - (10.0));
        local tics5  = math.abs(idealspacing - (5.0));
        local tics2  = math.abs(idealspacing - (2.0));
        local tics1  = math.abs(idealspacing - (1.0));

        if (tics10 < tics5 and tics10 < tics2 and tics10 < tics1) then
            gspacing = 10.0;
        elseif (tics5 < tics2 and tics5 < tics1) then
            gspacing = 5.0;
        elseif (tics2 < tics1) then
            gspacing = 2.0;
        else
            gspacing = 1.0;
        end
    elseif (units == 1) then								-- pi and degree spacing for degrees
        label.symbol = "\u00B0";
	if (idealspacing<=(5*degToRad)) then
            gspacing=5*degToRad;
        elseif (idealspacing<=(15*degToRad)) then
            gspacing=15*degToRad;
        elseif (idealspacing<=(45*degToRad)) then
            gspacing=45*degToRad;
        elseif (idealspacing<=(90*degToRad)) then
            gspacing=90*degToRad;
        elseif (idealspacing<(360*degToRad)) then
            gspacing=180*degToRad;
        else
            local n=idealspacing/(360*degToRad);
            gspacing=n*(360*degToRad);
        end
    elseif (units == 2)	then									-- pi spacing for radians
        label.symbol = "\u03c0";
        if (idealspacing <= (math.pi/32.0)) then
            gspacing=math.pi/32.0;
            label.divisor = 32;
        elseif (idealspacing <= (math.pi/16.0)) then
            gspacing=math.pi/16.0;
            label.divisor = 16;
        elseif (idealspacing <= (math.pi/8.0)) then
            gspacing=math.pi/8.0;
            label.divisor = 8;
        elseif (idealspacing <= (math.pi/4.0)) then
            gspacing=math.pi/4.0;
            label.divisor = 4;
        elseif (idealspacing <= (math.pi/2.0)) then
            gspacing=math.pi/2.0;
            label.divisor = 2;
        elseif (idealspacing < (2.0*math.pi)) then
            gspacing=math.pi;
        else
            local n=idealspacing/(math.pi*2.0);
            gspacing=n*(math.pi*2.0);
        end
    elseif (units == 3) then										-- pi/3 spacing for radians
        label.symbol = "\u03c0";
        if (idealspacing <= (math.pi/36.0)) then
            gspacing=math.pi/36.0;
            label.divisor = 36;
        elseif (idealspacing <= (math.pi/18.0)) then
            gspacing=math.pi/18.0;
            label.divisor = 18;
        elseif (idealspacing <= (math.pi/6.0)) then
            gspacing=math.pi/6.0;
            label.divisor = 6;
        elseif (idealspacing <= (math.pi/3.0)) then
            gspacing=math.pi/3.0;
            label.divisor = 3;
        elseif (idealspacing <= (math.pi/2.0)) then
            gspacing=math.pi/2.0;
            label.divisor = 2;
        elseif (idealspacing < (2.0*math.pi)) then
            gspacing=math.pi;
        else
            local n=idealspacing/(math.pi*2.0);
            gspacing=n*(math.pi*2.0);
        end
    end

    -- apply the magnitude to spacing
    gspacing =gspacing * magnitude;
    return gspacing;
end

function drawAxisTic(mat, ctx, type, pos, col, thickness, ticHalf)
    local point;
    ctx.color = col
    ctx.size = thickness
    if (type == 0) then
        point = transformCoord(mat, pos, 0);
        ctx:moveTo(point.x, point.y-ticHalf);
        ctx:lineTo(point.x, point.y+ticHalf);
	
    else
        point = transformCoord(mat, 0, pos);
        ctx:moveTo(point.x-ticHalf, point.y);
        ctx:lineTo(point.x+ticHalf, point.y);
    end
end

function drawTicLabel(mat, ctx, type, pos, dist, units, label, clrStr, ticHalf)

end


--common master functions


function paint(ctx, cnvs, scrn, wrld, TranMat, InvTranMat)

	-- Clear screen by setting canvas size
	--cnvs.width = cnvs.width;
	
	--Border
	--ctx.strokeRect(0.5, 0.5, (scrn.xmax - scrn.xmin)-1, (scrn.ymax - scrn.ymin)-1);
	
	-- Call generated Draw function
	drawShapes(TranMat, InvTranMat, wrld, ctx);
	
end
	  


function doFullUpdate(wrld, scrn, cnvs, ctx, TranMat, InvTranMat)	  
    ctx:init()
	updateInputs();
	computeGeometry();
	resetBounds(wrld, scrn, TranMat, InvTranMat);
	updateOutputs();
    resetDragHandles(TranMat)
	paint(ctx, cnvs, scrn, wrld, TranMat, InvTranMat);
end

function doMoveUpdate(wrld, scrn, cnvs, ctx, TranMat, InvTranMat)	  
    ctx:init()
	updateInputs();
	computeGeometry();
	updateOutputs();
    resetDragHandles(TranMat)
	paint(ctx, cnvs, scrn, wrld, TranMat, InvTranMat);
end

function doAnimationUpdate(wrld, scrn, cnvs, ctx, TranMat, InvTranMat)	  
    ctx:init()
	computeGeometry();
	updateOutputs();
    resetDragHandles(TranMat)
	paint(ctx, cnvs, scrn, wrld, TranMat, InvTranMat);
end

function onAnimate()
	doAnimationUpdate(commonGlobal.world, commonGlobal.screenBox, commonGlobal.canvas, commonGlobal.context, commonGlobal.world2Screen, commonGlobal.screen2World);
    platform.window:invalidate()

end

function onChangeView()
    commonGlobal.context:init()
	setMatrices(commonGlobal.world, commonGlobal.screenBox, commonGlobal.world2Screen, commonGlobal.screen2World)
        resetDragHandles(commonGlobal.world2Screen)
	paint(commonGlobal.context, commonGlobal.canvas, commonGlobal.screenBox, commonGlobal.world, commonGlobal.world2Screen, commonGlobal.screen2World)
        platform.window:invalidate()
end


function onDragHandleMove()
	doMoveUpdate(commonGlobal.world, commonGlobal.screenBox, commonGlobal.canvas, commonGlobal.context, commonGlobal.world2Screen, commonGlobal.screen2World);
    platform.window:invalidate()

end




function onPaint(gc)
	if commonGlobal ~= nil then
		local bb = commonGlobal.screenBox
		if bb ~= nil then
			gc:setColorRGB(unpack(BACKGROUND_COLOR))
			gc:fillRect(bb.xmin,bb.ymin,bb.xmax - bb.xmin, bb.ymax-bb.ymin)
			local objects = commonGlobal.context.objects
			if objects ~= nil then
				for i=1,#objects do
					local ob = objects[i]
					if ob.fill then
						gc:setColorRGB(unpack(ob.fillColor))
						if ob.type == "circle" then
							local r = ob.radius
							local x,y = ob[1],ob[2]
							gc:fillArc(x-r,y-r,2*r,2*r,0,360)
						elseif ob.type == "polyLine" then
							gc:fillPolygon(ob)
						end
					end
					local thickness,style = "thin","smooth"
					if ob.size ~= nil then
					    if ob.size > 4 then
						thickness = "thick"
					    elseif ob.size >2 then
						thickness = "medium"
					    end
					end
					if ob.style ~= nil then
						style = ob.style
					end
					gc:setPen(thickness,style)
					gc:setColorRGB(unpack(ob.color))
					if ob.type == "line" then
						gc:drawLine(ob[1],ob[2],ob[3],ob[4])
					elseif ob.type == "point" then  
						local r = ob.size
						local x,y = ob[1],ob[2]
						gc:fillArc(x-r,y-r,2*r,2*r,0,360)
					elseif ob.type == "circle" then
						local r = ob.radius
						local x,y = ob[1],ob[2]
						gc:drawArc(x-r,y-r,2*r,2*r,0,360)
					elseif ob.type == "string" then
						local x,y = ob[1],ob[2]
						local sw = gc:getStringWidth(ob.text)
						local sh = gc:getStringHeight(ob.text)
						if ob.horizontalAlignment == "right" then
							x = x-sw
						end
						gc:setFont("sansserif", "r", 9)
						gc:setColorRGB(0, 0, 0)
						gc:drawString(ob.text,x,y,ob.verticalAlignment)
					elseif ob.type == "polyLine" then
						if ob.clipped == nil then
							clip(ob,-32000,32000)
							ob.clipped = true
						end
						gc:drawPolyLine(ob)
					end
				end
			end
		end
	end
end
	
function onClickObject(selectObj)
    if components.focus ~= nil then
        if components.focus.setActive ~= nil then
            components.focus:setActive(false)
        end
   end
    components:setFocus(selectObj)
    if selectObj.onEnterKey ~= nil then
        selectObj:onEnterKey()
    end
    platform.window:invalidate()
end

function onClickPlayButton(selectObj)
     if selectObj:getData("state")=="play" then                              -- "Play" state
	local control = selectObj:getData("control")
	local delta = control.delta*DELTA_MULTIPLIER or 0.5
	local timerID = timer.performWithDelay(delta,control.increment,0)
	selectObj:setData("timerID",timerID)
     
        selectObj:setImage(image.new(mediaStopImage))                      -- change to reset state
        selectObj:setData("state", "reset")
    else                                                                        -- "Reset" state
    
	local timerID = selectObj:getData("timerID")
	timer.cancel(timerID)
        selectObj:setImage(image.new(mediaPlayImage))                       -- change to play state
        selectObj:setData("state", "play")
    end
    platform.window:invalidate()
end

function onClickControlUp(selectObj)
	if controlArray.visibleStart > 1 then
		controlArray.visibleStart = controlArray.visibleStart-1
		updateComponents()
		platform.window:invalidate()
	end
end

function onClickControlDown(selectObj)
	if not controlArray[#controlArray].objLabel.visible  then    -- can scroll
		controlArray.visibleStart = controlArray.visibleStart+1
		updateComponents()
		platform.window:invalidate()
	end
end

function stopDragging()
    if components.dragControl ~= nil then
        if components.dragControl.dragObject~=nil then                          -- we are dragging
            local x,y = components.dragControl.dragObject:getPos()
            components.dragControl.onMouseDown(x,y)                               -- will stop the dragging
            components.dragControl.onMouseUp(x,y)                               -- will stop the dragging
        end
    end
end

function onArrowLeft()
    if components.focus == objScreen then
	onArrowLeftScreen()
    else
	stopDragging()
	if components.focus ~= nil 
		and components.focus.getActive ~= nil 
		and components.focus:getActive() 
		and components.focus.onArrowLeft ~= nil then
	    components.focus:onArrowLeft() 
	    platform.window:invalidate()
	else
	    components:advanceFocus(-1)
	end
    end
    platform.window:invalidate()
end

function onArrowRight()
    if components.focus == objScreen then
	onArrowRightScreen()
    else
	stopDragging()
	local ff = components.focus
	if components.focus ~= nil 
		and components.focus.getActive ~= nil 
		and components.focus:getActive() 
		and components.focus.onArrowRight ~= nil then
	    components.focus:onArrowRight() 
	    platform.window:invalidate()
	else
	    components:advanceFocus(1)
	end
    end
    platform.window:invalidate()
end

function onArrowUp()
    if components.focus == objScreen then
	onArrowUpScreen()
    end
    platform.window:invalidate()
end

function onArrowDown()
    if components.focus == objScreen then
	onArrowDownScreen()
    end
    platform.window:invalidate()
end

function onTab (direction)
    stopDragging()
    if components.focus ~= nil then
        if components.focus.setActive ~= nil then
            components.focus:setActive(false)
        end
   end

    platform.window:invalidate()
    
    components:advanceFocus(direction)
end

function onLoseFocus()
 --   if theData ~= nil then
 --       theData:save()
 --   end
end

function onSliderChange()
    if haveSetupComponents ~= nil then
	doFullUpdate(commonGlobal.world, commonGlobal.screenBox, commonGlobal.canvas, commonGlobal.context, commonGlobal.world2Screen, commonGlobal.screen2World);
        platform.window:invalidate()
    end
end

function onCharInScreen(ch)  -- do zoom
    local wrld = commonGlobal.world;
    if ch == "+" then
        zoomWorld((wrld.xmin+wrld.xmax)/2, (wrld.ymin+wrld.ymax)/2,0.8);
        onChangeView();
    elseif ch == "-" then
        zoomWorld((wrld.xmin+wrld.xmax)/2, (wrld.ymin+wrld.ymax)/2,1.25);
        onChangeView();
    end
end

function onArrowUpScreen()  -- pan up
    panWorld(0,-0.1)
    onChangeView()
end

function onArrowDownScreen()  -- pan up
    panWorld(0,0.1)
    onChangeView()
end

function onArrowLeftScreen()  -- pan up
    panWorld(0.1,0)
    onChangeView()
end

function onArrowRightScreen()  -- pan up
    panWorld(-0.1,0)
    onChangeView()
end


function onCharIn(ch)
    if components.focus == objScreen then
	onCharInScreen(ch)
    elseif components.focus ~= nil and components.focus.onCharIn ~= nil then
        components.focus:onCharIn(ch) 
        platform.window:invalidate()
    end
end

function onGetFocus()

end

function onDeleteKey()
    if components.focus ~= nil then
        if components.focus.onDeleteKey ~= nil then
            components.focus:onDeleteKey()
        end
   end
end

function onBackspaceKey()
    if components.focus ~= nil then
        if components.focus.onBackspaceKey ~= nil then
            components.focus:onBackspaceKey()
        end
   end
end

function onEnterKey()
    if components.focus ~= nil then
        if components.focus.onEnterKey ~= nil then
            components.focus:onEnterKey()
        end
   end

end

function onAcceptValue(selectObj)
    local active = selectObj:getActive()
    if active then
	doFullUpdate(commonGlobal.world, commonGlobal.screenBox, commonGlobal.canvas, commonGlobal.context, commonGlobal.world2Screen, commonGlobal.screen2World);
    end
    selectObj:setActive(not active)                  

    platform.window:invalidate()

end




Context = class()

function Context:init()
	self.objects = {}
    self.lines = {}
    self.points = {}
    self.circles = {}
    self.color = {0,0,0}
    self.size = 1
end

function Context:moveTo(x,y)
    local line = {fillColor = self.fillColor, fill = self.fill, color = self.color, antialiasing = true, connect = true, size = self.size, type = "polyLine", style = self.style }
    line[1] = x
    line[2] = y
    self.objects[#self.objects+1]=line 
end 

function Context:lineTo(x,y)
    local nn = #self.objects
    self.objects[nn][#self.objects[nn]+1] = x
    self.objects[nn][#self.objects[nn]+1] = y
end

function Context:quadraticCurveTo(cpx,cpy,x,y)
    local nn = #self.objects
    local line = self.objects[nn]
    local x0 = line[#line-1]
    local y0 = line[#line]
    local dt = 1.0/CURVE_GRANULARITY
    for i=0, CURVE_GRANULARITY do
        local t = i*dt
        local xx=((t*cpx*2.0)+(t^2*cpx*(-2.0))+(x*t^2)+((1.0+(t*(-2.0))+t^2)*x0));
        local yy=((t*cpy*2.0)+(t^2*cpy*(-2.0))+(y*t^2)+((1.0+(t*(-2.0))+t^2)*y0));
        self:lineTo(xx,yy)
    end
end

function Context:submitMoveTo(mat, point)
	if isFinite(point.x) and isFinite(point.y) then
		local tpoint = transformCoord(mat, point.x, point.y);
		self:lineTo(tpoint.x, tpoint.y);
	end
end

function Context:submitLineTo(mat, point)
	if isFinite(point.x) and isFinite(point.y) then
		local tpoint = transformCoord(mat, point.x, point.y);
		self:lineTo(tpoint.x, tpoint.y);
	end
end

function Context:submitPolyLine(mat, points)
    for i = 1, #points do
        if (isFinite(points[i].x) and isFinite(points[i].y)) then
            local point = transformCoord(mat, points[i].x, points[i].y);
            self:lineTo(point.x, point.y);
        end
    end
end

function Context:submitPolyLineReverse(mat, points)
    for i = 1, #points do
       local point = points[#points-i+1];
       if (isFinite(point.x) and isFinite(point.y)) then
            local tpoint = transformCoord(mat, point.x, point.y);
            self:lineTo(tpoint.x, tpoint.y);
        end
    end
end

function Context:submitFill(color)
    self.fillColor = color;
    self.fill=true
end

function Context:clearFill()
    self.fillColor = nil;
    self.fill=nil
end



function Context:submitStroke(color, thickness,linestyle)
    self.color = color
    self.lineWidth = thickness
    self.size = thickness
    self.style = linestyle
end

function Context:submitBeginPath()
    local line = { color = self.color,fillColor = self.fillColor,fill = self.fill, antialiasing = true, connect = true, size = self.size ,type = "polyLine", style = self.style}
    self.objects[#self.objects+1]=line 
end

function Context:createAuxiliaryContext()
    local auxContext = Context()
    return auxContext
end

function applyAlpha(col,alpha)
    local faded_col = {}
    local fading_factor = 1-alpha
    for i = 1,3 do
        faded_col[i] = (255 - col[i]) * fading_factor + col[i]
    end
    return faded_col
end
        

-- Draw image data from a previously-created auxiliary canvas
function Context:drawAuxiliaryContext(aux, alpha)
    for i=1,#aux.objects do
	local col = aux.objects[i].color
        if col ~= nil then
            aux.objects[i].color = applyAlpha(col,alpha)
        end
	col = aux.objects[i].fillColor
        if fillColor ~= nil then
            aux.objects[i].fillColor = applyAlpha(col,alpha)
        end
        self.objects[#self.objects+1] = aux.objects[i]
    end
end



function load()

	--localiables
	commonGlobal = {}
	firstTimeIn = true
	
	commonGlobal.canvas = {width = 640, height=480}
	commonGlobal.context = Context()
	
	
	commonGlobal.screenBox = BoundingBox();
	commonGlobal.screenBox:extend(0,0);
	commonGlobal.screenBox:extend(commonGlobal.canvas.width, commonGlobal.canvas.height);
	commonGlobal.world = BoundingBox();
	commonGlobal.mouseScrn = Point(0.5, 0.5);
	commonGlobal.mouseWrld = Point(0, 0);
	commonGlobal.world2Screen = Tran2DMatrix();
	commonGlobal.screen2World = Tran2DMatrix();
        commonGlobal.axisIdealLabelCount = 10;

	commonGlobal.currentDraggable = 0;
	commonGlobal.dragStart = nil;
	commonGlobal.dragging = false;
    

	init();

end



--- TI environment stuff

platform.apilevel='2.0'


-- *****************
-- **** modules ****
-- *****************

-- #begin
-- #include "modules/Util.lua"
--------------------------------------------------------------------
-- Util library
--------------------------------------------------------------------
util = {}

function util.mouseTracker(obj, lx, ly, xform)
    local dx, dy = obj:getPos()
    local tx, ty = lx, ly
    local trans=xform
    if trans~=nil then
        tx, ty=trans:transformPoint(lx, ly)
    end
    dx = dx - tx
    dy = dy - ty
    return function(lx, ly)
        local tx,ty=lx,ly
        if trans~=nil then
            tx,ty=trans:transformPoint(lx, ly)
        end
        obj:setPos(tx + dx, ty + dy)
        platform.window:invalidate()
    end
end

--------------------------------------------------------------------
-- End util library
--------------------------------------------------------------------
-- #include "modules/On.lua"
--------------------------------------------------------------------
-- On module v0_1
--
-- This library supports adding and removing listeners for the on events
-- This is especially important when there are multiple controllers
-- interested in listening for the same events.  The listener mechanism
-- allows multiple users of the on events without conflict.
--------------------------------------------------------------------

on.listeners = {activate={}, arrowDown={}, arrowKey={}, arrowLeft={}, arrowRight={}, arrowUp={}, charIn={},
                backspaceKey={}, backtabKey={}, clearKey={}, contextMenu={}, copy={}, create={},
                createMathBox={}, cut={}, deactivate={}, deleteKey={}, destroy={}, enterKey={}, escapeKey={},
                getFocus={}, getSymbolList={}, grabDown={}, grabMove={}, grabUp={}, help={}, loseFocus={},
                mouseDown={}, mouseMove={}, mouseUp={}, paint={}, paste={}, resize={}, restore={},
                returnKey={}, rightMouseDown={}, rightMouseUp={}, save={}, tabKey={}, timer={}, varChange={}
}

on.PRIORITY_MAX=0
on.PRIORITY_MIN=100
on.PRIORITY_DEFAULT=(on.PRIORITY_MAX+on.PRIORITY_MIN)/2

function on.addListener(name, l, p, once)
    local plocal=p
    if p==nil then plocal=on.PRIORITY_DEFAULT end
    
    local listenerInfo = {key=name, listener = l, priority = plocal, oneShot = once}
    if on.listeners[name]==nil then on.listeners[name]={} end
    table.insert(on.listeners[name], listenerInfo)
    table.sort(on.listeners[name], function(A, B) return A.priority<B.priority end)
    return listenerInfo
end

function on.removeListener(listenerInfo)
    table.remove(on.listeners[listenerInfo.key], listenerInfo[value])
end

function on.fire(params)
    local listenerList = table.remove(params, 1)
    if listenerList~=nil then
        for _, listenerInfo in ipairs(listenerList) do
            listenerInfo.listener(unpack(params))
            if (listenerInfo.oneShot) then on.removeListener(listenerInfo) end
        end
    end
end

function on.activate() on.fire({on.listeners.activate}) end
function on.arrowDown() on.fire({on.listeners.arrowDown}) end
function on.arrowKey(key) on.fire({on.listeners.arrowKey, key}) end
function on.arrowLeft() on.fire({on.listeners.arrowLeft}) end
function on.arrowRight() on.fire({on.listeners.arrowRight}) end
function on.arrowUp() on.fire({on.listeners.arrowUp}) end
function on.charIn(char) on.fire({on.listeners.charIn, char}) end
function on.backspaceKey() on.fire({on.listeners.backspaceKey}) end
function on.backtabKey() on.fire({on.listeners.backtabKey}) end
function on.clearKey() on.fire({on.listeners.clearKey}) end
function on.contextMenu() on.fire({on.listeners.contextMenu}) end
function on.copy() on.fire({on.listeners.copy}) end
function on.create() on.fire({on.listeners.create}) end
function on.createMathBox() on.fire({on.listeners.createMathBox}) end
function on.cut() on.fire({on.listeners.cut}) end
function on.deactivate() on.fire({on.listeners.deactivate}) end
function on.deleteKey() on.fire({on.listeners.deleteKey}) end
function on.destroy() on.fire({on.listeners.destroy}) end
function on.enterKey() on.fire({on.listeners.enterKey}) end
function on.escapeKey() on.fire({on.listeners.escapeKey}) end
function on.getFocus() on.fire({on.listeners.getFocus}) end
function on.getSymbolList() on.fire({on.listeners.getSymbolList}) end
function on.grabDown(x, y) on.fire({on.listeners.grabDown, x, y}) end
function on.grabMove(x, y) on.fire({on.listeners.grabMove, x, y}) end
function on.grabUp(x, y) on.fire({on.listeners.grabUp, x, y}) end
function on.help() on.fire({on.listeners.help}) end
function on.loseFocus(gc) on.fire({on.listeners.loseFocus, gc}) end
function on.mouseDown(x, y) on.fire({on.listeners.mouseDown, x, y}) end
function on.mouseMove(x, y) on.fire({on.listeners.mouseMove, x, y}) end
function on.mouseUp(x, y) on.fire({on.listeners.mouseUp, x, y}) end
function on.paint(gc) on.fire({on.listeners.paint, gc}) end
function on.paste() on.fire({on.listeners.paste}) end
function on.resize(w,h) on.fire({on.listeners.resize, w, h}) end
function on.restore(state) on.fire({on.listeners.restore, state}) end
function on.returnKey() on.fire({on.listeners.returnKey}) end
function on.rightMouseDown(x,y) on.fire({on.listeners.rightMouseDown, x, y}) end
function on.rightMouseUp(x,y) on.fire({on.listeners.rightMouseUp, x, y}) end
function on.save() on.fire({on.listeners.save}) end
function on.tabKey() on.fire({on.listeners.tabKey}) end
function on.timer() on.fire({on.listeners.timer}) end
function on.varChange(varList) on.fire({on.listeners.varChange, varlist}) end

--------------------------------------------------------------------
-- End on module
--------------------------------------------------------------------
-- #include "modules/Timer.lua"
timer.timerID = 1
timer.activeTimers = {}
timer.pausedTimers = {}

timer.updateStart = function()
    nextFireTime = nil

    for _, timerObj in ipairs(timer.activeTimers) do
        if nextFireTime == nil or nextFireTime > timerObj.time then
            nextFireTime = timerObj.time
        end
    end

    if nextFireTime ~= nil then
        local delayTime = (nextFireTime - timer.getMilliSecCounter())/1000
        if (delayTime < .01) then
            delayTime = .01
        end

        timer.start(delayTime)
    else
        timer.stop()
    end
end

timer.fire = function()
    for i, timerObj in ipairs(timer.activeTimers) do
        if (timerObj.time <= timer.getMilliSecCounter()) then
            -- Fire timer and setup timer for next firing
            if (timerObj.number > 0) then
                -- decrement the number of times left to fire and remove when done
                timerObj.number = timerObj.number - 1
                if (timerObj.number == 0) then
                    table.remove(timer.activeTimers, i)
                end
            end

            
            table.remove(timer.activeTimers, i)
            timerObj.callback()
            timerObj.time = timer.getMilliSecCounter()+ 1000*timerObj.delay
             table.insert(timer.activeTimers, timerObj)
           
            if (timerObj.number == 0 and timerObj.onEnd) then
                timerObj.onEnd()
            end
            if (timerObj.invalidate) then
                platform.window:invalidate()
            end
        end
    end
    timer.updateStart()
end

--  PUBLIC METHODS
timer.performWithDelay = function (delay, callback, number, invalidate, onEnd)
    local timerObj = {
        time = timer.getMilliSecCounter() + delay,
        delay = delay,
        callback = callback,
        number = number or 1,
        invalidate = invalidate or false,
        onEnd = onEnd,
        id = timer.timerID
    }
    table.insert(timer.activeTimers, timerObj)

    timer.timerID = timer.timerID +1

    timer.updateStart()

    return timerObj.id
end

timer.cancel = function(timerID)
    for i, timerObj in ipairs(timer.activeTimers) do
        if (timerObj.id == timerID) then

            table.remove(timer.activeTimers, i)
        end
    end
    timer.updateStart()
end

timer.pause = function(timerID)
    for i, timerObj in ipairs(timer.activeTimers) do
        if (timerObj.id == timerID) then
            table.remove(timer.activeTimers, i)
            table.insert(timer.pausedTimers, timerObj)
        end
    end
    timer.updateStart()
end

timer.resume = function(timerID)
    for i, timerObj in ipairs(timer.pausedTimers) do
        if (timerObj.id == timerID) then
            table.remove(timer.pausedTimers, i)
            table.insert(timer.activeTimers, timerObj)
        end
    end
    timer.updateStart()
end

on.addListener("timer", function() timer.fire() end)

--------------------------------------------------------------------
-- End timer module
--------------------------------------------------------------------
-- #include "modules/Components.lua"
--------------------------------------------------------------------
-- Components library
--------------------------------------------------------------------

perf={
    step = true,
    drawingComponents = true,
    drawingImages = true,
    latestPaintTime = timer.getMilliSecCounter(),
    prevPaintTime = timer.getMilliSecCounter(),
    latestUpdateTime = timer.getMilliSecCounter(),
    prevUpdateTime = timer.getMilliSecCounter()
}

transform = {
    affine = function()

        local self = {
            matrix = {1,0,0, 0,1,0, 0,0,1},

            copy = function(self)
                local copyAffine = transform:affine()
                for i,v in ipairs(self.matrix) do
                    copyAffine.matrix[i] = v
                end
                return copyAffine
            end,
            
            scale = function(self, sx, sy)
                self.matrix[1] = self.matrix[1] * sx
                self.matrix[4] = self.matrix[4] * sx
                self.matrix[2] = self.matrix[2] * sy                
                self.matrix[5] = self.matrix[5] * sy
                return self
            end,

            translate = function(self, tx, ty)
                self.matrix[3] = self.matrix[3] + (tx * self.matrix[1]) + (ty * self.matrix[2])
                self.matrix[6] = self.matrix[6] + (tx * self.matrix[4]) + (ty * self.matrix[5])
                return self
            end,
            
            concatenate = function(self, tx)
                local m0 = self.matrix[1]
                local m1 = self.matrix[2]
                self.matrix[1] = (tx.matrix[1] * m0) + (tx.matrix[4] * m1)
                self.matrix[2] = (tx.matrix[2] * m0) + (tx.matrix[5] * m1)
                self.matrix[3] = self.matrix[3] + (tx.matrix[3] * m0) + (tx.matrix[6] * m1)
                
                m0 = self.matrix[4]
                m1 = self.matrix[5]
                self.matrix[4] = (tx.matrix[1] * m0) + (tx.matrix[4] * m1)
                self.matrix[5] = (tx.matrix[2] * m0) + (tx.matrix[5] * m1)
                self.matrix[6] = self.matrix[6] + (tx.matrix[3] * m0) + (tx.matrix[6] * m1)
                return self
            end,
            
            preConcatenate = function(self, tx)
                local m0 = self.matrix[1]
                local m1 = self.matrix[4]
                self.matrix[1] = (tx.matrix[1] * m0) + (tx.matrix[2] * m1)
                self.matrix[4] = (tx.matrix[4] * m0) + (tx.matrix[5] * m1)
                
                m0 = self.matrix[2]
                m1 = self.matrix[5]
                self.matrix[2] = (tx.matrix[1] * m0) + (tx.matrix[2] * m1)
                self.matrix[5] = (tx.matrix[4] * m0) + (tx.matrix[5] * m1)
                
                m0 = self.matrix[3]
                m1 = self.matrix[6]
                self.matrix[3] = (tx.matrix[1] * m0) + (tx.matrix[2] * m1) + tx.matrix[3]
                self.matrix[6] = (tx.matrix[4] * m0) + (tx.matrix[5] * m1) + tx.matrix[6]
                return self
            end,
            
            transformPoint = function(self, px, py)
                local x=nil
                local y=nil
                if px~=nil and py~=nil then
                    x = (px * self.matrix[1]) + (py * self.matrix[2]) + self.matrix[3]
                    y = (px * self.matrix[4]) + (py * self.matrix[5]) + self.matrix[6]
                end
                return x, y
            end,
           
            transformVector = function(self, vx, vy)
                local x=nil
                local y=nil
                if vx~=nil and vy~=nil then
                    x = (vx * self.matrix[1]) + (vy * self.matrix[2])
                    y = (vx * self.matrix[4]) + (vy * self.matrix[5])
                end
                return x, y
            end,

            getDeterminant = function(self)
                return (self.matrix[1]*self.matrix[5])-(self.matrix[2]*self.matrix[4])
            end,

            createInverse = function(self)
                local det=self:getDeterminant()
                local invAffine = transform:affine()
                invAffine.matrix[1]=self.matrix[5]/det
                invAffine.matrix[4]=-self.matrix[4]/det
                invAffine.matrix[2]=-self.matrix[2]/det
                invAffine.matrix[5]=self.matrix[1]/det
                invAffine.matrix[3]=((self.matrix[2]*self.matrix[6])-(self.matrix[5]*self.matrix[3]))/det
                invAffine.matrix[6]=((self.matrix[4]*self.matrix[3])-(self.matrix[1]*self.matrix[6]))/det
                return invAffine
            end,
         }
        return self
    end
}


components = {
    space = nil,
    children = {},
    trans = transform.affine(),
    focus = nil,
    focusColor = {r = 128, g = 128, b = 255, a = 0},
    focusPen = {thickness = "thick", style = "smooth"},
    
    DRAW_FIRST=0,
    DRAW_LAST=100,
    DRAW_DEFAULT=50,
}

components.remove = function(self, child)
    for k,v in ipairs(self.children) do
        if v == child then
            table.remove(self.children, k)
            local b=child.body
            if b~=nil then 
                self.space:removeBody(b)
            end
            local s=child.shape
            if s~=nil then
                self.space:removeShape(s)
                self.space:removeStaticShape(s)
            end
            break
        end
    end
    return self
end

components.setParent = function(self, p)
    if (self.parent ~= nil) then self.parent:remove(self) end
    self.parent = p
    table.insert(p.children, self)
    return self
end

components.getTransform = function(self)
    return self.trans
end

components.setTransform = function(self, t)
    self.trans = t;
    return self
end

components.advanceFocus = function(self, dir)
    local flist={}                                                              -- create a list of items that will accept focus
    for k,v in ipairs(self.children) do                                         -- for each item
        if v:acceptsFocus() then                                                -- item accepts focus
            table.insert(flist, v)                                              -- add it to the list
        end
    end
   
    table.sort(flist, function(A, B) return A.focusOrder<B.focusOrder end)      -- sort into focus order

    local currentIndex=0
    for k,v in ipairs(flist) do                                                 -- for each item in the sorted list
        if self.focus==v then                                                   -- this item currently has focus
            currentIndex=k                                                      -- save its index 
        end
     end

    if currentIndex>0 then                                                      -- valid current index
        if dir<0 then                                                           -- change focus to previous item
            currentIndex=currentIndex-1                                         -- decrement index
        elseif dir>0 then                                                       -- change focus to next item
            currentIndex=currentIndex+1                                         -- increment index
        end
        
        if currentIndex<1 then                                                  -- past start of list
            currentIndex=#flist                                                 -- wrap to end of list
        elseif currentIndex>#flist then                                         -- reached end of list
            currentIndex=1                                                      -- wrap to start of list
        end
        
        self.focus=flist[currentIndex]                                          -- set new focus
        
    else                                                                        -- no current focus item
        if #flist>0 then                                                        -- there are items that accept focus
            self.focus=flist[1]                                                 -- set focus to the first item
        end
    end
    
    platform.window:invalidate()
end

components.hasFocus = function(self, child)
    return self.focus==child
end

components.setFocus = function(self, obj)
    if obj:acceptsFocus() then self.focus=obj end
end

components.updateDrawOrder = function(self)
    table.sort(self.children, function(A, B) return A.drawOrder<B.drawOrder end)
end

components.rect = class()

function components.rect:__init(wx, wy, ww, wh)
    self:initGeneric()

    self.w = ww or 1
    self.h = wh or 1
    self.x = wx
    self.y = wy

    return self
end

function components.rect:initGeneric()
    self.color = {r = 0, g = 0, b = 0, a = 0}
    self.filled = false

    self.penColor = {r = 0, g = 0, b = 0, a = 0}
    self.penStyle = {thickness = "thin", style = "smooth"}
    self.outlined = false

    self.draggable = {drag = false, onStart = nil, onMove = nil, onEnd = nil, bounds = nil}
    self.selectable = {select = false, onClick = nil}

    self.shape = nil -- There is only a shape if the body is controlled by the physics engine

    self.parent = nil
    self.children = {}
    self.useTransform = true
    self.xForm = transform.affine()
    
    self.focusOrder = -1
    
    self.data = {}
  
    self.drawOrder=components.DRAW_DEFAULT
    self.visible = true
    
    self:setParent(components)
end

function components.rect:setImage(image)
    self.image = image
    return self
end

function components.rect:getImage()
    return self.image
end

function components.rect:renderImage(gc)
    local cx, cy = self:getPos()
    local cw, ch = self:getSize()
    local x,y,w,h = self:toScreen(cx, cy, cw, ch)

    x=x+(w-self.image:width())/2                                    -- center the image
    y=y+(h-self.image:height())/2
    
    gc:drawImage(self:getImage(), x, y)
end

function components.rect:rotatePoint(cx, cy, ox, oy, rot)
    cx = cx - ox
    cy = cy - oy
    local nx = cx * math.cos(rot) - cy * math.sin(rot)
    local ny = math.sin(rot) * cx + math.cos(rot) * cy
    nx = nx + ox
    ny = ny + oy

    return nx, ny
end

function components.rect:render(gc, xform, hasFocus)
	if self.visible then
		local points = {}
		local cx, cy = self:getPos()
		local cw, ch = self:getSize()
		local x,y,w,h = self:toScreen(cx, cy, cw, ch)

		points[1], points[2] = x, y
		points[3], points[4] = x, y+h
		points[5], points[6] = x+w, y+h
		points[7], points[8] = x+w, y
		points[9], points[10] = points[1], points[2]

		if hasFocus then
			gc:setColorRGB(components.focusColor.r, components.focusColor.g, components.focusColor.b)
			gc:setPen(components.focusPen.thickness, components.focusPen.style)
			gc:drawPolyLine(points)
		end
	    
		if self:getFilled() then
			local color = self:getFillColor()
			gc:setColorRGB(color.r, color.g, color.b)
			gc:fillPolygon(points)
		end
	    
		if (self:getImage() ~= nil) then
			self:renderImage(gc)
		end

		if self:getOutlined() then
			local color = self:getPenColor()
			gc:setColorRGB(color.r, color.g, color.b)
			local style=self:getPenStyle()
			gc:setPen(style.thickness, style.style)
			gc:drawPolyLine(points)
		end
	end
end

function components.rect:setUseTransform(use)
    self.useTransform=use
    return self
end

function components.rect:getUseTransform()
    return self.useTransform
end

function components.rect:setXForm(xForm)
    self.xForm = xForm
end

function components.rect:toScreen(cx, cy, cw, ch)
--    x = cx + self.xForm.matrix[3]
--    y = cy + self.xForm.matrix[6]
--    return x,y, cw, ch

    local x, y = self.xForm:transformPoint(cx, cy)
    
    local w=cw
    local h=ch
    if cw~=nil and ch~=nil then
        w,h=self.xForm:transformVector(cw, ch)
    elseif cw~=nil then
        w,h=self.xForm:transformVector(cw, cw)
    end
    
    return x, y, w, h
end

function components.rect:hit(tx, ty)
    local x,y = self:getPos()
    local sx, sy, sw, sh = self:toScreen(x, y, self.w, self.h)
    return tx >= sx and tx <= sx+sw and ty >= sy and ty <= sy+sh
end

function components.rect:setParent(p)
    return components.setParent(self, p)
end
function components.rect:remove(child)
    return components.remove(self, child)
end



function components.rect:getPos()
    return self.x, self.y
end

function components.rect:setPos(wx, wy)
    self.x, self.y = wx,wy
    return self
end

function components.rect:getSize()
    return self.w,self.h
end

function components.rect:setSize(ww, wh)
    self.w = ww
    self.h = wh
    return self
end


function components.rect:getFillColor()
    return self.color
end

function components.rect:setFillColor(red, green, blue, alpha)
    self.color = {r = red, g = green, b = blue, a = alpha}
    platform.window:invalidate()
    return self
end

function components.rect:getFilled()
    return self.filled
end

function components.rect:setFilled(fill)
    self.filled = fill
    platform.window:invalidate()
    return self
end

function components.rect:getPenColor()
    return self.penColor
end

function components.rect:setPenColor(red, green, blue, alpha)
    self.penColor = {r = red, g = green, b = blue, a = alpha}
    platform.window:invalidate()
    return self
end

function components.rect:getPenStyle()
    return self.penStyle
end

function components.rect:setPenStyle(penThickness, penStyle)
    self.penStyle = {thickness = penThickness, style = penStyle}
    platform.window:invalidate()
    return self
end

function components.rect:getOutlined()
    return self.outlined
end

function components.rect:setOutlined(outlined)
    self.outlined = outlined
    platform.window:invalidate()
    return self
end

function components.rect:setDraggable(d)
    self.draggable.drag = d.drag or false
    self.draggable.onStart = d.onStart or nil
    self.draggable.onMove = d.onMove or nil
    self.draggable.onEnd = d.onEnd or nil
    self.draggable.bounds = d.bounds or nil
    return self
end

function components.rect:getDraggable()
    return self.draggable
end

function components.rect:setSelectable(d)
	if self.selectable == nil then
		self.selectable = {}
	end
    self.selectable.select = d.select or false
    self.selectable.onClick = d.onClick or nil
    return self
end

function components.rect:getSelectable()
    return self.selectable
end

function components.rect:setFocusOrder(order)
    self.focusOrder=order
    return self
end

function components.rect.getFocusOrder()
    return self.focusOrder
end

function components.rect:acceptsFocus()
    return self.focusOrder>=0
end

function components.rect:setData(key, value)
    if self.data==nil then
        self.data={}
    end
    self.data[key]=value
    return self
end

function components.rect:getData(key)
    local value=nil
    if self.data~=nil then
        value=self.data[key]
    end
    return value
end

function components.rect:setDrawOrder(order)
    self.drawOrder=order
    return self
end

function components.rect:getDrawOrder()
    return self.drawOrder
end

components.circle = class(components.rect)

function components.circle:__init(wx, wy, wr)
    self:initGeneric()

    self.r = wr or 1
    self:setPos(wx,wy)

    return self
end

function components.circle:setImage(image)
    self.image = image
    self.r = self.image:height()/2
    return self
end

function components.circle:render(gc, xform, hasFocus)
	if self.visible then
		if (self:getImage() ~= nil) then
			self:renderImage(gc)
		else
			local cx, cy = self:getPos()
			local cr = self:getSize()
			cx,cy = cx-cr,cy-cr   -- want top left corner
			local x,y,r = self:toScreen(cx, cy, cr)

			if self:getFilled() then
				local color = self:getFillColor()
				gc:setColorRGB(color.r, color.g, color.b)
				gc:fillArc(x, y, 2*r, 2*r, 0, 360)
			end
			if self:getOutlined() then
				local color = self:getPenColor()
				gc:setColorRGB(color.r, color.g, color.b)
				local style=self:getPenStyle()
				gc:setPen(style.thickness, style.style)        
				gc:drawArc(x, y, 2*r, 2*r, 0, 360)
			end
		end
	end
end

function components.circle:getPos()
    return self.x,self.y
end

function components.circle:setPos(wx, wy)
    self.x, self.y = wx,wy
    return self
end

function components.circle:getSize()
    return self.r
end

function components.circle:setSize(wr)
    self.r = wr
    return self
end

function components.circle:hit(tx, ty)
    local x,y = self:getPos()
    local sx, sy, sr = self:toScreen(x, y, self.r)
    return ((tx - sx)^2 + (ty - sy)^2) <= sr^2
end


components.text = class(components.rect)

function components.text:__init(wx, wy, ww, wh, t, halign, valign)
    self:initGeneric()

    self.text = t
    self.hAlignment=halign or "left"
    self.vAlignment=valign or "top"
    
    self.w = ww
    self.h = wh
    self:setPos(wx,wy)

    self.textColor = {r = 0, g = 0, b = 0, a = 0}
    self.textStyle = {family = "serif", style = "r", size=7}

    return self
end

function components.text:getText()
    return self.text
end

function components.text:setText(t)
    self.text = t
end

function components.text:getHorizontalAlignment()
    return self.hAlignment
end

function components.text:setHorizontalAlignment(a)
    self.hAlignment = a
end

function components.text:getVerticalAlignment()
    return self.vAlignment
end

function components.text:setVerticalAlignment(a)
    self.vAlignment = a
end

function components.text:getTextColor()
    return self.textColor
end

function components.text:setTextColor(red, green, blue, alpha)
    self.textColor = {r = red, g = green, b = blue, a = alpha}
    platform.window:invalidate()
    return self
end

function components.text:getTextStyle()
    return self.textStyle
end

function components.text:setTextStyle(textFamily, textStyle, textSize)
    self.textStyle = {family = textFamily, style = textStyle, size=textSize}
    platform.window:invalidate()
    return self
end

function components.text:render(gc, xform, hasFocus)
	if self.visible then
		local cx, cy = self:getPos()                                                -- get text location
		local cw, ch = self:getSize()                                               -- get text size
		local x,y,w,h = self:toScreen(cx, cy, cw, ch)                               -- tranfrom to screen coordinates

		local color = self:getTextColor()
		gc:setColorRGB(color.r, color.g, color.b)                                   -- set text color

		local ts=self:getTextStyle()
		gc:setFont(ts.family, ts.style, ts.size)                                    -- set text font

		local sw=gc:getStringWidth(self:getText())                                  -- get text extent
		local sh=gc:getStringHeight(self:getText())

		local halign=self:getHorizontalAlignment()                                  -- get text horizontal alignment
		if halign=="middle" then
			x=x+(w/2)-sw/2
		elseif halign=="right" then
			x=x+w-sw
		end
	    
		local valign=self:getVerticalAlignment()                                    -- get text vertical alignment
		if valign=="middle" then
			y=y+(h/2)-sh/2
		elseif valign=="bottom" then
			y=y+h-sh
		end

		gc:drawString(self:getText(), x, y, "top")
	end
end


components.drawSpace = function (gc)
    components:drawChildren(components, gc, components:getTransform())
end
components.drawChildren = function (self, c, gc, xForm)
    for _, bodyShape in ipairs(c.children) do
        local trans=xForm
        if bodyShape:getUseTransform() == false then                -- not using the transform
            trans = transform.affine()                              -- set the transform to identity
        end
        
        local hasFocus=(bodyShape==self.focus)
        
        bodyShape:setXForm(trans)
        bodyShape:render(gc, trans, hasFocus)

        components:drawChildren(bodyShape, gc, xForm:copy():translate(bodyShape:getPos()))
    end
end

components.dragControl = {
    downX=0,
    downY=0,
    dragObject=nil,
    buttonDownDrag=false,
    buttonDownThreshold=1,
    
    clear = function()
        components.dragControl.dragObject=nil
        components.dragControl.buttonDownDrag=false
    end,
    
    onMouseDown = function(wx, wy)
        if components.dragControl.dragObject==nil then                          -- we are not dragging
            components.dragControl.dragObject=components.dragControl.dragHitCheck(components.children, wx, wy) -- look for draggable object
            if components.dragControl.dragObject~=nil then                      -- we have an object to drag
                cursor.set("drag grab")                                         -- start drag
                components.dragControl.buttonDownDrag=false                     -- initially not a button down drag
                components.dragControl.downX=wx                                 -- save the mouse coordinates
                components.dragControl.downY=wy    
                if (components.dragControl.dragObject:getDraggable().onStart ~= nil) then components.dragControl.dragObject:getDraggable().onStart(components.dragControl.dragObject) end
                platform.window:invalidate()
            end
        else                                                                    -- we are dragging
            if (components.dragControl.dragObject:getDraggable().onEnd ~= nil) then components.dragControl.dragObject:getDraggable().onEnd(components.dragControl.dragObject) end
            components.dragControl.dragObject=nil                               -- drop the drag object
            cursor.set("pointer")
                platform.window:invalidate()
        end

    end,
    
    onMouseMove = function(wx, wy)
        if components.dragControl.dragObject==nil then                          -- we are not dragging
            if components.dragControl.dragHitCheck(components.children, wx, wy) ~= nil then -- over a draggable object
                cursor.set("hand open")                                         -- set cursor to open hand
            end        
        else                                                                    -- we are dragging
            cursor.set("drag grab")                                             -- set drag cursor
            local vx,vy = wx-components.dragControl.downX, wy-components.dragControl.downY -- get drag vector
            local dist=vx*vx+vy*vy                                               -- drag distance
            if dist>components.dragControl.buttonDownThreshold^2 then             -- distance exceeded our threshold
                components.dragControl.buttonDownDrag=true                      -- we are doing a button down drag
            end
            
            local x,y=wx,wy
            if components.dragControl.dragObject:getUseTransform() then
                itrans=components.dragControl.dragObject.xForm:createInverse()
                x,y=itrans:transformPoint(wx, wy)
            end
            if (components.dragControl.dragObject:getDraggable().onMove ~= nil) then
                components.dragControl.dragObject:getDraggable().onMove(components.dragControl.dragObject, x, y)
            else
                components.dragControl.dragObject:setPos(x, y)
            end
            platform.window:invalidate()
        end
    end,
    
    onMouseUp = function(wx, wy)
        if components.dragControl.dragObject==nil then                          -- we are not dragging
        else                                                                    -- we are dragging
            if components.dragControl.buttonDownDrag then                       -- button down drag
                if (components.dragControl.dragObject:getDraggable().onEnd ~= nil) then components.dragControl.dragObject:getDraggable().onEnd(components.dragControl.dragObject) end
                components.dragControl.dragObject=nil                           -- end drag
            else
                cursor.set("drag grab")                                         -- start button up drag
                platform.window:invalidate()
            end
        end

    end,
    
    onGrabDown = function(wx, wy)
            components.dragControl.dragObject=nil
            components.dragControl.onMouseDown(wx, wy)								-- automatically pick up the object
            components.dragControl.onMouseUp(wx, wy)
    end,
    
    onGrabUp = function(wx, wy)
            components.dragControl.onMouseUp(wx, wy)
    end,

    draggableCheck = function(wx, wy)
        if components.dragControl.dragHitCheck(components.children, wx, wy) ~= nil then
            cursor.set("hand open")
        end
    end,

    dragHitCheck = function(parent, wx, wy)
        for _, bodyShape in ipairs(parent) do
            if bodyShape:getDraggable().drag and bodyShape:hit(wx, wy, "drag") then

                return bodyShape
            end
        end
    end,

    dragCheck = function(wx, wy)
        local dragShape = components.dragControl.dragHitCheck(components.children, wx, wy)

        if dragShape ~= nil then
            cursor.set("drag grab")
            if (dragShape:getDraggable().onStart ~= nil) then dragShape:getDraggable().onStart(dragShape) end
            local itrans=nil
            if dragShape:getUseTransform() then
                itrans=dragShape.xForm:createInverse()
            end
            local moveListener = on.addListener("mouseMove", util.mouseTracker(dragShape, wx, wy, itrans))

            local upListener = on.addListener("mouseUp",
                function(wx, wy)
                    on.removeListener(moveListener)
                    if (dragShape:getDraggable().onEnd ~= nil) then dragShape:getDraggable().onEnd(dragShape) end
                end,
                on.PRIORITY_DEFAULT,
                true)
        end
    end
}

components.selectControl = {
    selectObject=nil,
    
    selectableCheck = function(wx, wy)
        if components.selectControl.selectHitCheck(components.children, wx, wy) ~= nil then
            cursor.set("hand pointer")
        end
    end,

    selectHitCheck = function(parent, wx, wy)
        for _, bodyShape in ipairs(parent) do
            if bodyShape:getSelectable().select and bodyShape:hit(wx, wy, "select") then
                return bodyShape
            end
        end
    end,

    selectCheck = function(wx, wy)
        selectObject = components.selectControl.selectHitCheck(components.children, wx, wy)

        if selectObject ~= nil then
            if (selectObject:getSelectable().onClick ~= nil) then selectObject:getSelectable().onClick(selectObject) end
        end
    end
}

on.addListener("paint", function(gc) components.drawSpace(gc) end)
--on.addListener("mouseDown", function(wx, wy) components.dragControl.dragCheck(wx, wy) end)
--on.addListener("mouseMove", function(wx, wy) components.dragControl.draggableCheck(wx, wy) end)
on.addListener("mouseDown", function(wx, wy) components.selectControl.selectCheck(wx, wy) end)
on.addListener("mouseMove", function(wx, wy) components.selectControl.selectableCheck(wx, wy) end)
on.addListener("mouseDown", function(wx, wy) components.dragControl.onMouseDown(wx, wy) end)
on.addListener("mouseMove", function(wx, wy) components.dragControl.onMouseMove(wx, wy) end)
on.addListener("mouseUp", function(wx, wy) components.dragControl.onMouseUp(wx, wy) end)
on.addListener("mouseDown", function(wx, wy) if theLayers ~= nil then theLayers:onMouseDown(wx, wy) end end)
on.addListener("grabDown", function(wx, wy) components.dragControl.onGrabDown(wx, wy) end)
on.addListener("grabUp", function(wx, wy) components.dragControl.onGrabUp(wx, wy) end)

on.addListener("arrowLeft", onArrowLeft)
on.addListener("arrowRight", onArrowRight)
on.addListener("arrowUp", onArrowUp)
on.addListener("arrowDown", onArrowDown)
on.addListener("backtabKey", function() onTab(-1) end)
on.addListener("deleteKey", onDeleteKey)
on.addListener("backspaceKey", onBackspaceKey)
on.addListener("enterKey", onEnterKey)
on.addListener("charIn", onCharIn)
on.addListener("mouseDown", function(wx, wy) mouseX=wx mouseY=wy if wx==0 and wy==0 then onEnterKey() end end)
on.addListener("mouseMove", function(wx, wy) mouseX=wx mouseY=wy end)
on.addListener("mouseUp", function(wx, wy) mouseX=wx mouseY=wy end)
on.addListener("tabKey",  function() onTab(1) end)
on.addListener("sliderChange", onSliderChange)
on.addListener("loseFocus", onLoseFocus)
on.addListener("getFocus", onGetFocus)



--------------------------------------------------------------------
-- End components library
--------------------------------------------------------------------
-- #include "modules/Slider.lua"
--------------------------------------------------------------------
--------------------------------------------------------------------
-- Components library slider control
--------------------------------------------------------------------
components.slider = class(components.text)

function components.slider:__init(wx, wy, ww, wh)
    self:initGeneric()

    self.numberOfIncrements = 100
    self.draggable = {drag = true, onStart = nil, onMove = components.slider.onDragMove, onEnd = nil, bounds = nil}

    self.w = ww or 1
    self.h = wh or 1
    self:setPos(wx,wy)

    self:setRange(0, 100)
    self.divisions=4

    self.textColor = {r = 0, g = 0, b = 0, a = 0}
    self.textStyle = {family = "serif", style = "r", size=7}

    self.borderWidth=2
    self.borderHeight=2
    
    self.pointerX=0
    self.pointerY=0
    self.pointerWidth=0
    self.pointerHeight=0
    
    self.slotWidth=0
    
    self.stringFormat="%.2f"
    
    self.variableName=nil
    
    self.active=false
    
    self.increment = 1
    
    return self
end

function components.slider:hit(tx, ty, hitType)
    local hitResult=false
    local x,y = self:getPos()
    local sx, sy, sw, sh = self:toScreen(x, y, self.w, self.h)
    if hitType=="drag" then
        hitResult = (tx >= self.pointerX and tx <= self.pointerX+self.pointerWidth and ty >= self.pointerY and ty <= self.pointerY+self.pointerHeight)
    elseif hitType=="select" then
        hitResult = (tx >= sx and tx <= sx+sw and ty >= sy and ty <= sy+sh)
    end
    return hitResult
end

function components.slider:render(gc, xform, hasFocus)
	if self.visible then
		local points = {}
		local cx, cy = self:getPos()
		local cw, ch = self:getSize()
		local x,y,w,h = self:toScreen(cx, cy, cw, ch)
		local isActive=self:getActive()
	    
		if (hasFocus and (isActive==false)) then
			gc:setColorRGB(components.focusColor.r, components.focusColor.g, components.focusColor.b)
			gc:setPen(components.focusPen.thickness, components.focusPen.style)
			gc:drawRect(x, y, w, h)
		end
	    
		if self:getFilled() then
			local color = self:getFillColor()
			gc:setColorRGB(color.r, color.g, color.b)
			gc:fillRect(x, y, w, h)
		end
	    
		if (self:getImage() ~= nil) then
			self:renderImage(gc)
		end

		local ts=self:getTextStyle()
		gc:setFont(ts.family, ts.style, ts.size)                                    -- set text font

		local currentTextW=gc:getStringWidth(self:getValueAsString())               -- get extent of current value string
		local currentTextH=gc:getStringHeight(self:getValueAsString())
		local rangeMin,rangeMax=self:getRangeAsString()
		local maxTextW=gc:getStringWidth(rangeMax)                                  -- get extent of max value string

		local tx=x+w-currentTextW-self.borderWidth                                  -- text right justified in slider rect
		local ty=y+((h-currentTextH)/2)
		gc:setColorRGB(self.textColor.r,self.textColor.g,self.textColor.b)
		gc:drawString(self:getValueAsString(), tx, ty, "top")

		local pointerH=h/2                                                          -- pointer dimensions
		local pointerW=pointerH/2
	    
		local leftX=x+self.borderWidth+(pointerW/2)                                 -- get drawing coordinates
		local rightX=x+w-self.borderWidth-maxTextW-self.borderWidth-(pointerW/2)
		local midY=y+(h/2)
	    
		local slotH=pointerH/4                                 
		local slotW=rightX-leftX

		local outlineColor=self:getPenColor()                                       -- get color info for drawing the slider
		local outlineStyle=self:getPenStyle()
		local fillColor=self:getFillColor()

		gc:setColorRGB(fillColor.r, fillColor.g, fillColor.b)                       -- draw slider
		gc:fillRect(leftX, midY-slotH/2, slotW, slotH)
		gc:setColorRGB(outlineColor.r, outlineColor.g, outlineColor.b)
		gc:setPen(outlineStyle.thickness, outlineStyle.style)
		gc:drawRect(leftX, midY-slotH/2, slotW, slotH)

		local pointerX=leftX+((self:getValue()-rangeMin)/(rangeMax-rangeMin)*slotW)            -- draw pointer
		local points = {}
		points[1], points[2] = pointerX-(pointerW/2), midY-(pointerH/2)
		points[3], points[4] = pointerX+(pointerW/2), midY-(pointerH/2)
		points[5], points[6] = pointerX+(pointerW/2), midY+(pointerH/4)
		points[7], points[8] = pointerX, midY+(pointerH/2)
		points[9], points[10] = pointerX-(pointerW/2), midY+(pointerH/4)
		points[11], points[12] = points[1], points[2]

		if (hasFocus and (isActive==true)) then
			gc:setColorRGB(components.focusColor.r, components.focusColor.g, components.focusColor.b)
			gc:setPen(components.focusPen.thickness, components.focusPen.style)
			gc:drawPolyLine(points)
		end

		gc:setColorRGB(fillColor.r, fillColor.g, fillColor.b)
		gc:fillPolygon(points)
		gc:setColorRGB(outlineColor.r, outlineColor.g, outlineColor.b)
		gc:setPen(outlineStyle.thickness, outlineStyle.style)
		gc:drawPolyLine(points)

		local ticIncrement=slotW/self:getDivisions()                                -- draw tic marks
		local ticTop=midY+(slotH/2)+slotH
		local ticBottom=ticTop+slotH
	    
		gc:setColorRGB(outlineColor.r, outlineColor.g, outlineColor.b)
		gc:setPen(outlineStyle.thickness, outlineStyle.style)
		for ticX=leftX, rightX, ticIncrement do
			gc:drawLine(ticX, ticTop, ticX, ticBottom)
		end

		if self:getOutlined() then
			local color = self:getPenColor()
			gc:setColorRGB(color.r, color.g, color.b)
			local style=self:getPenStyle()
			gc:setPen(style.thickness, style.style)
			gc:drawRect(x, y, w, h)
		end
	    
		self.pointerX=points[1]                                                     -- save pointer info for drag hit test
		self.pointerY=points[2]
		self.pointerWidth=points[3]-points[1]
		self.pointerHeight=points[8]-points[2]

		self.slotWidth=slotW   
	end                                                     -- save slot pixel width for drag value calculation
end

function components.slider:onDragMove(x, y)
    if self.slotWidth>0 then                                                    -- valid slot width
        local dx=x-self.pointerX                                                -- pixel movement in the x direction
        local rangeMin,rangeMax=self:getRange()
        local dv=(dx/self.slotWidth)*(rangeMax-rangeMin)                        -- change in value
        self:setValue(self:getValue()+dv)                                       -- set new value
    end
end

function components.slider:setRange(minVal, maxVal)
    self.minValue=minVal or 0
    self.maxValue=maxVal or 1

    if self.maxValue<self.minValue then
        self.minValue,self.maxValue=self.maxValue,self.minValue
    end
    
    self.value=self.minValue
    
    self.increment = (self.maxValue - self.minValue) / self.numberOfIncrements
    return self
end

function components.slider:getRange()
    return self.minValue, self.maxValue
end

function components.slider:getRangeAsString()
    return string.format(self.stringFormat, self.minValue), string.format(self.stringFormat, self.maxValue)
end

function components.slider:setDivisions(div)
    if div>0 then self.divisions=div end
    return self
end

function components.slider:getDivisions()
    return self.divisions
end

function components.slider:setValue(v)
    if v<self.minValue then v=self.minValue end                                 -- clip value to slider range
    if v>self.maxValue then v=self.maxValue end
    if self.value~=v then                                                       -- value has changed
        self.value=v                                                            -- set the new value
        if self.variableName~=nil then                                          -- variable name has been set
            var.store(self.variableName, self.value)                            -- store the new value in the variable
        end
        on.fire({on.listeners.sliderChange, self})                          -- fire off a message
    end
    return self
end

function components.slider:getValue()
    return self.value
end

function components.slider:getValueAsString()
    return string.format(self.stringFormat, self.value)
end

function components.slider:setVariableName(name)
    self.variableName=name
    return self
end

function components.slider:getVariableName()
    return self.variableName
end

function components.slider:setActive(active)
    self.active=active
    return self
end

function components.slider:getActive()
    return self.active
end

function components.slider:onArrowLeft()
    self:setValue(self:getValue() - self.increment)
end

function components.slider:onArrowRight()
    self:setValue(self:getValue() + self.increment)
end

--------------------------------------------------------------------
-- End components library slider control
--------------------------------------------------------------------
--------------------------------------------------------------------
-- End components library slider control
--------------------------------------------------------------------
-- #end





components.textEdit = class(components.text)



function components.textEdit:__init(wx, wy, ww, wh, t, halign, valign)
    self:initGeneric()

    self.text = t

    self.w = ww or 1
    self.h = wh or 1
    self:setPos(wx, wy)
    self.hAlignment=halign or "right"
    self.vAlignment=valign or "middle"

    self.textColor = {r = 0, g = 0, b = 0, a = 0}
    self.textStyle = {family = "serif", style = "r", size=7}

    self.borderWidth=2
    self.borderHeight=2
    self.preTextGap = 0
    
    self.xOffset = 0    
    
    self.stringFormat="%d"
    
    self.maxChars = 15
    self.selectionColor = {212,212,212}
    
    self.active=false
    
    return self
end



function components.textEdit:getText()
    return self.text
end

function components.textEdit:splitText()
    local text = self:getText()
    local str1,str2,str3 = "","",""
    if self.selectionStart and self.selectionFinish then
        if self.selectionStart >0 then
            str1 = string.sub(text,1,self.selectionStart)
        end
        str2 = string.sub(text,self.selectionStart+1,self.selectionFinish)
        if self.selectionFinish < #text then
            str3 = string.sub(text,self.selectionFinish+1,#text)
        end

    elseif self.cursorLocation then
        if self.cursorLocation >0 then
            str1 = string.sub(text,1,self.cursorLocation)
        end
        if self.cursorLocation <#text then
            str3 = string.sub(text,self.cursorLocation+1,#text)
        end
    else
        str1 = text
    end
    
    return str1, str2, str3

end

function components.textEdit:render(gc, xform, hasFocus)
	if self.visible then
		local points = {}
		local cx, cy = self:getPos()
		local cw, ch = self:getSize()
		local x,y,w,h = self:toScreen(cx, cy, cw, ch)
		local isActive=self:getActive()
	    
		self:putCursorLocationInView(gc)
	    
		if (hasFocus and (isActive==false)) then
			gc:setColorRGB(components.focusColor.r, components.focusColor.g, components.focusColor.b)
			gc:setPen(components.focusPen.thickness, components.focusPen.style)
			gc:drawRect(x, y, w, h)
		end
	    
		if self:getFilled() then
			local color = self:getFillColor()
			gc:setColorRGB(color.r, color.g, color.b)
			gc:fillRect(x, y, w, h)
		end
	    
		if (self:getImage() ~= nil) then
			self:renderImage(gc)
		end

		local ts=self:getTextStyle()
		gc:setFont(ts.family, ts.style, ts.size)                                    -- set text font

		local currentTextW=gc:getStringWidth(self:getText())               -- get extent of current value string
		local currentTextH=gc:getStringHeight(self:getText())

		local tx=x+self.borderWidth+self.preTextGap -  self.xOffset                                -- text right justified in textEdit rect
		local ty=y+((h-currentTextH)/2)
		gc:setColorRGB(self.textColor.r,self.textColor.g,self.textColor.b)
	    
		gc:clipRect("set",x,y,w,h)
		local str1, str2, str3 = self:splitText()
		if #str1>0 then
			gc:drawString(str1, tx, ty, self.vAlignment)
			tx = tx + gc:getStringWidth(str1)
		end
		if #str2>0 then
			local sw = gc:getStringWidth(str2)
			local sh = gc:getStringHeight(str2)
	        
			gc:setColorRGB(unpack(self.selectionColor))
			local hy = ty
			gc:fillRect(tx,hy, sw, currentTextH)
	        
			gc:setColorRGB(self.textColor.r,self.textColor.g,self.textColor.b)

			gc:drawString(str2, tx, ty, self.vAlignment)
			tx = tx + sw
		end
		if #str3>0 then
			gc:drawString(str3, tx, ty, self.vAlignment)
			tx = tx + gc:getStringWidth(str3)
		end
	    
		if self.cursorLocation ~= nil then
			local cursor_x = x+self.borderWidth+self.preTextGap - self.xOffset
			if self.cursorLocation >0 then
				cursor_x = cursor_x + gc:getStringWidth(string.sub(self.text,1,self.cursorLocation))
			end
			gc:setColorRGB(0,0,0)
		   gc:setPen("thin","smooth")
		   gc:drawLine(cursor_x,ty,cursor_x,ty+h)
	   end
	            
	            
	    
	    
		gc:clipRect("reset")
	    

		if self:getOutlined() then
			local color = self:getPenColor()
			gc:setColorRGB(color.r, color.g, color.b)
			local style=self:getPenStyle()
			gc:setPen(style.thickness, style.style)
			gc:drawRect(x, y, w, h)
		end
	end	    
end

function components.textEdit:setSelection(s0,s1)
    self.selectionStart = s0
    self.selectionFinish = s1
    self.cursorLocation = nil
end

function components.textEdit:setCursorLocation(c)
    self.selectionStart = nil
    self.selectionFinish = nil
    self.cursorLocation = c
end

function components.textEdit:putLocationInView(gc,c)
    local cw, ch = self:getSize()
    if self.didDelete ~= nil then
		self.didDelete = nil
		local dd = gc:getStringWidth(string.sub(self.text,1,#self.text))
		if dd<cw+self.xOffset-self.borderWidth-self.preTextGap then
			self.xOffset = dd-cw+self.borderWidth+self.preTextGap+1
			if self.xOffset <0 then self.xOffset = 0 end
		end
	end

    if c>0 then
        local dx = gc:getStringWidth(string.sub(self.text,1,c))
        if dx < self.xOffset then 
			self.xOffset = dx
		elseif dx > cw+self.xOffset-self.borderWidth-self.preTextGap then
			self.xOffset = dx-cw+self.borderWidth+self.preTextGap+1
		end
    end
end

function components.textEdit:putCursorLocationInView(gc)
	if self.cursorLocation == nil then
		self:putLocationInView(gc,1)	
	else
		self:putLocationInView(gc,self.cursorLocation)	
	end
end



function components.textEdit:setActive(active)
    self.active=active
    if active == true then
        local text = self.text
        if (#text>0) then
            self:setSelection(0,#text)
        else
            self:setCursorLocation(0)
        end
    else
        self.selectionStart = nil
        self.selectionFinish = nil
        self.cursorLocation = nil
    end
    return self
end

function components.textEdit:getActive()
    return self.active
end

function components.textEdit:onCharIn(ch)
    if not self:getActive() then
        self:setActive(true)
    end
    local str1,str2,str3 = self:splitText()
    local text = str1 .. ch .. str3
    if #text <= self.maxChars then        
        self.text = text
        self:setCursorLocation(#str1+1)
    end
end

function components.textEdit:onArrowLeft()
    if self.selectionStart ~= nil then
        self:setCursorLocation(self.selectionStart)
    elseif self.cursorLocation ~=nil and self.cursorLocation > 0 then
        self:setCursorLocation(self.cursorLocation - 1)
    else
        self:setCursorLocation(0)
    end
end

function components.textEdit:onArrowRight()
    if self.selectionFinish ~= nil then
        self:setCursorLocation(self.selectionFinish)
    elseif self.cursorLocation ~=nil and self.cursorLocation < #self.text then
        self:setCursorLocation(self.cursorLocation + 1)
    else
        self:setCursorLocation(#self.text)
    end
end

function components.textEdit:onBackspaceKey(ch)
    if self:getActive() then
        local str1,str2,str3 = self:splitText()
        if #str2 > 0 then   -- backspacing a selection
            self.text = str1 .. str3
            self:setCursorLocation(#str1)
        elseif #str1 > 0 then
            str1 = string.sub(str1,1,#str1-1)
            self.text = str1 .. str3
            self:setCursorLocation(#str1)
        end
        self.didDelete = true
        platform.window:invalidate()

    end
end

function components.textEdit:onDeleteKey(ch)
    if self:getActive() then
        local str1,str2,str3 = self:splitText()
        if #str2 > 0 then   -- deleting a selection
            self.text = str1 .. str3
            self:setCursorLocation(#str1)
        elseif #str3 > 0 then
            str3 = string.sub(str3,2,#str3)
            self.text = str1 .. str3
            self:setCursorLocation(#str1)
        end
        self.didDelete = true
        platform.window:invalidate()
    end
end

function components.textEdit:onEnterKey()
        self:setActive(not self:getActive())                  -- toggle mass slider active state
        platform.window:invalidate()
end


-- end components.textEdit

MIN_WIDTH_FOR_LARGE_COMPONENTS = 400

DRAW_ORDER_SCREEN_RECT = 60
DRAW_ORDER_CONTROL_POINT = 60
DRAW_ORDER_STATUS = 80
DRAW_ORDER_CONTROL=90
DRAW_ORDER_CALLOUT=95
DRAW_ORDER_MESSAGE=100

-- ****************
-- **** images ****
-- ****************




mediaPauseLarge =
-- #begin
-- #include "res/mediaPause48.lua"
"0\000\000\0000\000\000\000\000\000\000\000`\000\000\000\016\000\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\015\128\211\128\211\128\211\128\211\128\211\128\211\128\211\128\211\128\211\128\211\128\211\128\210\128\015\128\000\000\000\000\000\000\000\000\000\000\000\000\015\128\211\128\211\128\211\128\211\128\211\128\211\128\211\128\211\128\211\128\211\128\211\128\210\128\015\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\015\128\211\1284\1418\182Y\186Y\186Y\186Y\1869\1829\1829\1829\1828\178U\149\211\128\000\000\000\000\000\000\000\000\000\000\015\128\211\1284\1418\182Y\186Y\186Y\186Y\1869\1829\1829\1829\1828\178U\149\211\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\211\128\020\141\251\210\252\214\252\210\251\210\251\206\219\206\219\206\219\206\219\202\219\202\219\202\219\202U\145\210\128\000\000\000\000\000\000\000\000\211\128\020\141\251\210\252\214\252\210\251\210\251\206\219\206\219\206\219\206\219\202\219\202\219\202\219\202U\145\210\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\211\128\248\173\251\210\219\206\219\206\219\202\219\202\187\202\187\198\187\198\187\198\187\198\187\198\187\198\248\169\211\128\000\000\000\000\000\000\000\000\211\128\248\173\251\210\219\206\219\206\219\202\219\202\187\202\187\198\187\198\187\198\187\198\187\198\187\198\248\169\211\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\211\1288\178\219\206\219\202\187\198\187\198\187\194\155\194\154\194\154\190\154\190\155\190\155\190\155\190\024\174\211\128\000\000\000\000\000\000\000\000\211\1288\178\219\206\219\202\187\198\187\198\187\194\155\194\154\194\154\190\154\190\155\190\155\190\155\190\024\174\211\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\211\128\024\174\187\198\187\194\154\194\154\190\154\190z\186z\186z\186z\186z\182z\182z\182\248\165\211\128\000\000\000\000\000\000\000\000\211\128\024\174\187\198\187\194\154\194\154\190\154\190z\186z\186z\186z\186z\182z\182z\182\248\165\211\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\211\128\248\169\154\194\154\190z\186z\186z\182z\182Z\182Z\178Z\178Z\178Z\178Z\178\248\161\211\128\000\000\000\000\000\000\000\000\211\128\248\169\154\194\154\190z\186z\186z\182z\182Z\182Z\178Z\178Z\178Z\178Z\178\248\161\211\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\211\128\248\165z\186z\186Z\182Z\178Z\178Z\174Z\174:\174:\174:\170:\170Z\170\216\157\211\128\000\000\000\000\000\000\000\000\211\128\248\165z\186z\186Z\182Z\178Z\178Z\174Z\174:\174:\174:\170:\170Z\170\216\157\211\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\211\128\216\161Z\182Z\178Z\178:\174:\1709\1709\170:\166:\166:\166:\166:\166\216\153\211\128\000\000\000\000\000\000\000\000\211\128\216\161Z\182Z\178Z\178:\174:\1709\1709\170:\166:\166:\166:\166:\166\216\153\211\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\211\128\216\161Z\178:\1749\1709\166\025\166\025\162\025\162\025\162\025\158\025\158\026\158\026\158\184\149\211\128\000\000\000\000\000\000\000\000\211\128\216\161Z\178:\1749\1709\166\025\166\025\162\025\162\025\162\025\158\025\158\026\158\026\158\184\149\211\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\211\128\216\157:\1709\166\025\166\025\162\025\158\249\157\249\157\249\153\249\153\249\153\249\153\250\153\184\145\244\128\000\000\000\000\000\000\000\000\211\128\216\157:\1709\166\025\166\025\162\025\158\249\157\249\157\249\153\249\153\249\153\249\153\250\153\184\145\244\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\211\128\184\153\025\166\025\162\025\158\249\157\249\153\249\149\249\149\217\149\217\145\217\145\217\145\250\145\152\141\244\128\000\000\000\000\000\000\000\000\211\128\184\153\025\166\025\162\025\158\249\157\249\153\249\149\249\149\217\149\217\145\217\145\217\145\250\145\152\141\244\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\211\128\184\149\025\162\249\157\249\153\249\149\217\149\217\145\217\141\217\141\217\141\217\141\217\141\217\141\152\137\244\128\000\000\000\000\000\000\000\000\211\128\184\149\025\162\249\157\249\153\249\149\217\149\217\145\217\141\217\141\217\141\217\141\217\141\217\141\152\137\244\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\211\128\184\145\249\153\249\149\217\149\217\145\217\141\185\141\185\137\185\137\185\133\185\133\185\133\185\133\152\133\244\128\000\000\000\000\000\000\000\000\211\128\184\145\249\153\249\149\217\149\217\145\217\141\185\141\185\137\185\137\185\133\185\133\185\133\185\133\152\133\244\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\243\128\184\145\249\149\217\145\217\141\185\137\185\137\185\133\185\133\153\129\185\129\185\129\185\129\186\129x\129\244\128\000\000\000\000\000\000\000\000\243\128\184\145\249\149\217\145\217\141\185\137\185\137\185\133\185\133\153\129\185\129\185\129\185\129\186\129x\129\244\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\244\128\152\141\217\145\217\141\185\137\185\133\185\129\153\129\153\129\153\129\185\129\186\129\186\129\186\129x\129\020\129\000\000\000\000\000\000\000\000\244\128\152\141\217\145\217\141\185\137\185\133\185\129\153\129\153\129\153\129\185\129\186\129\186\129\186\129x\129\020\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\244\128\152\137\217\141\185\137\185\133\153\129\153\129\153\129\185\129\186\129\186\129\186\129\186\129\186\129\152\129\020\129\000\000\000\000\000\000\000\000\244\128\152\137\217\141\185\137\185\133\153\129\153\129\153\129\185\129\186\129\186\129\186\129\186\129\186\129\152\129\020\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\244\128\152\137\185\137\185\133\185\129\185\129\185\129\186\129\186\129\186\129\186\129\186\129\218\129\218\129\153\129\021\129\000\000\000\000\000\000\000\000\244\128\152\137\185\137\185\133\185\129\185\129\185\129\186\129\186\129\186\129\186\129\186\129\218\129\218\129\153\129\021\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\244\128\152\133\185\133\185\129\185\129\186\129\186\129\186\129\186\129\186\129\218\129\218\129\219\129\219\129\153\129\021\129\000\000\000\000\000\000\000\000\244\128\152\133\185\133\185\129\185\129\186\129\186\129\186\129\186\129\186\129\218\129\218\129\219\129\219\129\153\129\021\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\244\128\152\129\186\129\186\129\186\129\186\129\186\129\186\129\218\129\219\129\219\129\219\129\219\129\219\129\153\129\021\129\000\000\000\000\000\000\000\000\244\128\152\129\186\129\186\129\186\129\186\129\186\129\186\129\218\129\219\129\219\129\219\129\219\129\219\129\153\129\021\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\020\129\152\129\186\129\186\129\186\129\186\129\218\129\219\129\219\129\219\129\219\129\219\129\251\129\251\129\185\129\021\129\000\000\000\000\000\000\000\000\020\129\152\129\186\129\186\129\186\129\186\129\218\129\219\129\219\129\219\129\219\129\219\129\251\129\251\129\185\129\021\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\020\129\153\129\186\129\218\129\218\129\219\129\219\129\219\129\219\129\219\129\251\129\251\129\252\129\252\129\186\129\021\129\000\000\000\000\000\000\000\000\020\129\153\129\186\129\218\129\218\129\219\129\219\129\219\129\219\129\219\129\251\129\251\129\252\129\252\129\186\129\021\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\021\129\153\129\218\129\219\129\219\129\219\129\219\129\219\129\251\129\251\129\252\129\252\129\252\129\252\129\186\1295\129\000\000\000\000\000\000\000\000\021\129\153\129\218\129\219\129\219\129\219\129\219\129\219\129\251\129\251\129\252\129\252\129\252\129\252\129\186\1295\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\021\129\153\129\219\129\219\129\219\129\219\129\251\129\251\129\252\129\252\129\252\129\252\129\252\129\028\130\186\1296\129\000\000\000\000\000\000\000\000\021\129\153\129\219\129\219\129\219\129\219\129\251\129\251\129\252\129\252\129\252\129\252\129\252\129\028\130\186\1296\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\021\129\185\129\219\129\219\129\251\129\252\129\252\129\252\129\252\129\252\129\028\130\028\130\029\130\029\130\219\1296\129\000\000\000\000\000\000\000\000\021\129\185\129\219\129\219\129\251\129\252\129\252\129\252\129\252\129\252\129\028\130\028\130\029\130\029\130\219\1296\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\021\129\186\129\251\129\252\129\252\129\252\129\252\129\252\129\028\130\028\130\029\130\029\130\029\130\029\130\219\1296\129\000\000\000\000\000\000\000\000\021\129\186\129\251\129\252\129\252\129\252\129\252\129\252\129\028\130\028\130\029\130\029\130\029\130\029\130\219\1296\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\0005\129\186\129\252\129\252\129\252\129\252\129\028\130\028\130\029\130\029\130\029\130\029\130\029\130=\130\219\1296\129\000\000\000\000\000\000\000\0005\129\186\129\252\129\252\129\252\129\252\129\028\130\028\130\029\130\029\130\029\130\029\130\029\130=\130\219\1296\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\0006\129\186\129\252\129\252\129\028\130\028\130\029\130\029\130\029\130\029\130\029\130=\130>\130>\130\251\129V\129\000\000\000\000\000\000\000\0006\129\186\129\252\129\252\129\028\130\028\130\029\130\029\130\029\130\029\130\029\130=\130>\130>\130\251\129V\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\0006\129\219\129\028\130\029\130\029\130\029\130\029\130\029\130\029\130=\130>\130>\130>\130>\130\252\129W\129\000\000\000\000\000\000\000\0006\129\219\129\028\130\029\130\029\130\029\130\029\130\029\130\029\130=\130>\130>\130>\130>\130\252\129W\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\0006\129\219\129\029\130\029\130\029\130\029\130=\130=\130>\130>\130>\130>\130>\130^\130\252\129W\129\000\000\000\000\000\000\000\0006\129\219\129\029\130\029\130\029\130\029\130=\130=\130>\130>\130>\130>\130>\130^\130\252\129W\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\0006\129\219\129\029\130\029\130=\130=\130>\130>\130>\130>\130>\130^\130^\130_\130\252\129W\129\000\000\000\000\000\000\000\0006\129\219\129\029\130\029\130=\130=\130>\130>\130>\130>\130>\130^\130^\130_\130\252\129W\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\0006\129\251\129=\130=\130>\130>\130>\130>\130>\130^\130_\130_\130_\130_\130\028\130W\129\000\000\000\000\000\000\000\0006\129\251\129=\130=\130>\130>\130>\130>\130>\130^\130_\130_\130_\130_\130\028\130W\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000V\129\252\129>\130>\130>\130>\130>\130^\130_\130_\130_\130_\130_\130\127\130\029\130W\129\000\000\000\000\000\000\000\000V\129\252\129>\130>\130>\130>\130>\130^\130_\130_\130_\130_\130_\130\127\130\029\130W\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000W\129\252\129>\130>\130^\130^\130_\130_\130_\130_\130_\130\127\130\127\130\127\130\029\130w\129\000\000\000\000\000\000\000\000W\129\252\129>\130>\130^\130^\130_\130_\130_\130_\130_\130\127\130\127\130\127\130\029\130w\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000W\129\028\130^\130^\130_\130_\130_\130_\130_\130\127\130\127\130\127\130\127\130\127\130\029\130x\129\000\000\000\000\000\000\000\000W\129\028\130^\130^\130_\130_\130_\130_\130_\130\127\130\127\130\127\130\127\130\127\130\029\130x\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000W\129\029\130_\130_\130_\130_\130_\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\029\130x\129\000\000\000\000\000\000\000\000W\129\029\130_\130_\130_\130_\130_\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\029\130x\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000W\129=\134\127\134_\130_\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\029\130x\129\000\000\000\000\000\000\000\000W\129=\134\127\134_\130_\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\029\130x\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000w\129=\138\159\138\127\134\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\029\130x\129\000\000\000\000\000\000\000\000w\129=\138\159\138\127\134\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\029\130x\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000w\129<\138\159\142\159\138\127\134\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\252\129\152\129\000\000\000\000\000\000\000\000w\129<\138\159\142\159\138\127\134\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\252\129\152\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000x\129\153\129\159\146\159\142\159\138\127\134\127\134\127\130\127\130\127\130\127\130\127\130\127\130_\130\153\129\153\129\000\000\000\000\000\000\000\000x\129\153\129\159\146\159\142\159\138\127\134\127\134\127\130\127\130\127\130\127\130\127\130\127\130_\130\153\129\153\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000U\129x\129\185\129<\138=\138=\134=\134=\134=\130=\130=\130=\130\028\130\185\129\153\129\000\000\000\000\000\000\000\000\000\000U\129x\129\185\129<\138=\138=\134=\134=\134=\130=\130=\130=\130\028\130\185\129\153\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000U\129x\129x\129x\129\152\129\152\129\152\129\152\129\152\129\152\129\152\129\153\129\152\129\031\128\000\000\000\000\000\000\000\000\000\000\000\000U\129x\129x\129x\129\152\129\152\129\152\129\152\129\152\129\152\129\152\129\153\129\152\129\031\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
-- #end

mediaPauseSmall =
-- #begin
-- #include "res/mediaPause24.lua"
"\024\000\000\000\024\000\000\000\000\000\000\0000\000\000\000\016\000\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\210\128\211\128\211\128\211\128\211\128\211\128\211\128\210\128\178\128\000\000\000\000\210\128\211\128\211\128\211\128\211\128\211\128\211\128\211\128\178\128\000\000\000\000\000\000\000\000\211\128\150\153y\190y\186Y\186Y\186Y\182\183\161\211\128\000\000\000\000\211\128\182\161y\190y\186Y\186Y\186Y\182\150\153\211\128\000\000\000\000\000\000\000\000\211\1289\182\219\206\219\202\187\198\187\198\187\198z\182\211\128\211\128s\128\211\128y\186\219\206\219\202\187\198\187\198\187\1989\174\211\128\000\000\000\000\000\000\000\000\211\128\024\174\187\194\154\190\154\186z\186z\186:\174\211\128\175\128q\128\211\1289\178\155\194\154\190z\186z\186z\186\025\166\211\128\000\000\000\000\000\000\000\000\211\128\248\165z\182Z\178Z\174Z\174Z\170\025\162\211\128\175\128q\128\211\128\025\170z\182Z\178Z\174Z\174Z\170\249\157\211\128\000\000\000\000\000\000\000\000\211\128\216\157:\170:\166\025\162\026\158\026\158\249\153\244\128\175\128q\128\211\128\249\161:\170:\166\025\162\026\158\026\158\217\149\244\128\000\000\000\000\000\000\000\000\211\128\216\149\025\158\249\153\249\149\217\145\217\145\217\141\244\128\175\128q\128\211\128\217\153\025\158\249\153\249\149\217\145\249\145\185\141\244\128\000\000\000\000\000\000\000\000\243\128\184\145\217\149\217\141\185\137\185\133\185\133\185\133\244\128\181\128\241\128\243\128\217\145\217\145\217\141\185\137\185\133\185\133\153\129\244\128\000\000\000\000\000\000\000\000\244\128\184\137\185\137\185\129\185\129\185\129\186\129\186\129\020\129\181\128\241\128\244\128\185\137\185\137\185\129\185\129\185\129\186\129\153\129\020\129\000\000\000\000\000\000\000\000\244\128\153\133\185\129\186\129\186\129\186\129\218\129\186\129\021\129\181\128\241\128\244\128\185\133\185\129\186\129\186\129\186\129\218\129\186\129\021\129\000\000\000\000\000\000\000\000\020\129\153\129\186\129\186\129\218\129\219\129\219\129\219\129\021\129\181\128\241\128\020\129\185\129\186\129\186\129\218\129\219\129\219\129\218\129\021\129\000\000\000\000\000\000\000\000\021\129\185\129\219\129\219\129\219\129\251\129\252\129\251\1295\129\181\128\245\128\021\129\186\129\219\129\219\129\219\129\251\129\252\129\219\1295\129\000\000\000\000\000\000\000\000\021\129\186\129\219\129\251\129\252\129\252\129\252\129\252\1296\129\181\128\245\128\021\129\219\129\219\129\251\129\252\129\252\129\252\129\251\1296\129\000\000\000\000\000\000\000\000\021\129\219\129\252\129\252\129\028\130\029\130\029\130\029\1306\129\181\128\245\128\021\129\219\129\252\129\252\129\028\130\029\130\029\130\252\1296\129\000\000\000\000\000\000\000\0006\129\219\129\028\130\029\130\029\130\029\130=\130\029\130V\129\181\128\245\1286\129\252\129\028\130\029\130\029\130\029\130=\130\029\130V\129\000\000\000\000\000\000\000\0006\129\252\129\029\130\029\130=\130>\130>\130>\130W\129U\129\245\1286\129\028\130\029\130\029\130=\130>\130>\130\029\130W\129\000\000\000\000\000\000\000\0006\129\028\130>\130>\130>\130^\130_\130^\130W\129U\129\245\1286\129\029\130>\130>\130>\130^\130_\130>\130W\129\000\000\000\000\000\000\000\000V\129\029\130>\130^\130_\130_\130_\130_\130w\129U\129\245\128W\129>\130>\130^\130_\130_\130_\130>\130w\129\000\000\000\000\000\000\000\000W\129=\130_\130_\130\127\130\127\130\127\130_\130x\129U\129U\129W\129^\130_\130_\130\127\130\127\130\127\130>\130x\129\000\000\000\000\000\000\000\000W\129^\138\127\134\127\130\127\130\127\130\127\130_\130x\129\153\129\247\128W\129\127\138\127\134\127\130\127\130\127\130\127\130>\130x\129\000\000\000\000\000\000\000\000w\129\251\133~\138_\134_\130_\130^\130\251\129\152\129\000\000\000\000w\129\251\137~\138_\134_\130_\130^\130\219\129\152\129\000\000\000\000\000\000\000\000W\129x\129x\129x\129x\129\152\129\153\129\153\129\153\129\000\000\000\000x\129x\129x\129x\129x\129\152\129\153\129\152\129x\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
-- #end

mediaPlayLarge =
-- #begin
-- #include "res/mediaPlay32.lua"
" \000\000\000 \000\000\000\000\000\000\000@\000\000\000\016\000\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\213\128\146\128\146\128\178\128\212\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\212\128\244\128\219\206<\219y\190\243\128\212\128\182\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\146\128y\190\028\219\251\210\219\206\219\206\182\157\178\128\246\128\000\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\146\128\186\198\252\210\187\202\186\194\154\194\155\194Y\182\020\137\180\128\247\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\146\128\154\194\219\202\187\194\154\190z\186Z\182Z\178Z\178\183\153\211\128\213\128u\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\146\128z\190\187\198\154\190z\186Z\182Z\1749\170\025\166\025\166\025\1625\137\211\128\247\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\146\128z\186\187\194z\186Z\182Z\1749\170\025\166\025\162\249\157\249\153\249\149\184\141\244\128\245\128\247\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\146\128z\182\155\190z\182Z\174:\170\025\166\025\162\249\153\217\149\217\145\185\141\185\141\185\137V\129\244\128\022\129\018\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\178\128Z\178\155\186Z\178:\170:\166\025\162\249\153\249\149\217\145\185\141\185\137\153\133\153\129\153\129\153\1295\129\244\1288\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\178\128Z\178{\182Z\174:\166\026\162\025\158\249\149\217\145\217\141\185\133\153\129\153\129\153\129\185\129\186\129\186\129\152\129\021\129\022\129\023\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\179\128Z\174{\178Z\170:\162\026\158\250\149\217\145\217\141\185\133\185\129\153\129\185\129\185\129\186\129\186\129\186\129\219\129\219\129W\129\021\1299\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\179\128Z\170{\174:\166:\162\026\154\250\149\218\141\217\137\185\129\185\129\186\129\186\129\186\129\186\129\218\129\218\129\219\129\219\129\252\129\219\1296\1297\129S\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\211\128Z\170{\170:\162\026\158\026\150\250\145\218\137\186\133\186\129\186\129\186\129\186\129\218\129\218\129\219\129\219\129\219\129\219\129\252\129\252\129\029\130x\129\021\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\211\128[\166[\170;\162\026\154\026\150\250\141\218\133\186\129\186\129\186\129\218\129\219\129\219\129\219\129\219\129\251\129\251\129\252\129\252\129\252\129\029\130\028\130\021\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\211\128[\166[\166;\158\027\154\027\146\250\137\218\133\218\129\218\129\219\129\219\129\219\129\219\129\251\129\252\129\252\129\252\129\252\129\252\129\028\130\029\130\029\130\021\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\212\128[\166{\166;\158\027\150\027\142\251\137\219\129\219\129\219\129\219\129\219\129\251\129\252\129\252\129\252\129\252\129\028\130\028\130\029\130\029\130>\130\153\1296\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\244\128[\166|\166;\158;\150\027\142\251\133\219\129\219\129\251\129\251\129\252\129\252\129\252\129\252\129\028\130\029\130\029\130\029\130>\130\029\130w\129U\129\151\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\244\128\092\166|\166\092\154<\150\027\142\251\133\251\129\252\129\252\129\252\129\252\129\252\129\028\130\029\130\029\130\029\130=\130>\130\186\1296\129x\129\031\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\244\128|\166|\166\092\154<\150<\142\028\134\252\129\252\129\252\129\028\130\028\130\029\130\029\130\029\130=\130>\130\028\130W\129Y\129{\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\245\128|\166}\166\092\154\092\150<\142\028\134\028\130\028\130\029\130\029\130\029\130\029\130=\130>\130>\130\153\1296\129\155\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\245\128\157\166\157\166}\158]\150=\142=\138\029\130\029\130\029\130\029\130=\130>\130_\130\219\129W\129y\129[\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\021\129\157\166\157\166}\158}\150]\146=\138=\130\029\130=\130>\130^\130=\130x\129Y\129\155\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\021\129\189\166\190\170\157\158}\154}\146]\142>\134>\130>\130_\130\186\129W\129\155\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\021\129\190\170\190\170\158\162\158\154~\150~\142^\138_\130\252\129X\129y\129[\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\022\129\222\170\222\174\190\166\158\158\158\154\159\146\127\142\152\129v\129\155\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\0006\129\157\166\255\178\223\166\223\162\223\158\027\142W\129\153\129\031\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000y\129x\129\255\178\031\179\158\162\153\129y\129\188\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000z\1297\129W\1297\129\154\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
-- #end

mediaPlaySmall =
-- #begin
-- #include "res/mediaPlay16.lua"
"\016\000\000\000\016\000\000\000\000\000\000\000 \000\000\000\016\000\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\178\128U\145\211\128\180\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\020\133}\231\219\206v\149\180\128\181\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\020\137\252\210\154\190z\186\249\161\244\128\246\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\020\137\219\198Z\178:\166\026\158\250\149w\133\245\128\025\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\020\133\155\190:\170\026\158\217\145\185\137\185\129\153\1296\129\022\129\149\128\000\000\000\000\000\000\000\000\000\000\020\133\155\178:\162\250\149\217\137\185\129\186\129\218\129\219\129\185\1296\1297\129\000\000\000\000\000\000\000\0005\133|\174\026\154\250\141\186\129\186\129\219\129\219\129\219\129\252\129\029\1306\129\000\000\000\000\000\000\000\0005\133|\170\027\150\251\133\219\129\219\129\251\129\252\129\252\129\029\130>\1306\129\000\000\000\000\000\000\000\000V\133\156\166;\146\251\133\251\129\252\129\252\129\029\130>\130\219\129W\129X\129\000\000\000\000\000\000\000\000V\133\157\166<\146\028\134\028\130\029\130>\130=\130\153\129X\129\185\129\000\000\000\000\000\000\000\000\000\000w\133\190\170]\150=\134>\130^\130\219\129x\129\156\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000w\133\222\174\158\154\127\142\029\130x\129{\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000w\133?\187\223\162\218\137y\129\219\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\0007\129\185\137w\129z\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"

-- #end

mediaStopLarge =
-- #begin
-- #include "res/mediaStop32.lua"
" \000\000\000 \000\000\000\000\000\000\000@\000\000\000\016\000\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\146\128\178\128\146\128\178\128\178\128\178\128\178\128\178\128\178\128\178\128\178\128\178\128\178\128\178\128\210\128\178\128\178\128\178\128\178\128\178\128\178\128\178\128\178\128\210\128\210\128\178\128\211\128\143\128\000\000\000\000\000\000\000\000\178\128\020\137\028\219]\223\028\219\252\210\219\206\219\202\187\198\155\190z\186z\182Z\178Z\1749\1749\170:\166\025\166\026\162\025\162\026\162\026\158\026\158\026\158:\158:\1585\133\211\128\000\000\000\000\000\000\000\000\146\128\219\206\028\219\219\206\187\202\186\198\154\194z\186Z\182Y\1789\1749\170\025\166\025\166\249\161\249\157\249\153\217\153\217\149\217\149\217\149\217\149\217\149\217\149\249\149\250\149\026\150\211\128\000\000\000\000\000\000\000\000\146\128\028\215\251\210\219\202\187\198\154\194z\186Z\182Y\1789\174\025\170\025\166\249\161\249\157\249\153\216\153\216\149\216\145\216\145\185\145\185\141\185\141\217\141\217\141\217\141\217\141\027\146\211\128\000\000\000\000\000\000\000\000\146\128\252\210\219\206\187\198\154\194z\190z\182Z\1789\1749\170\025\166\249\161\249\157\217\153\216\149\216\145\184\145\184\141\184\137\185\137\185\137\185\137\185\133\185\133\185\137\217\137\250\141\243\128\000\000\000\000\000\000\000\000\146\128\252\206\219\202\187\194\154\190z\186Z\178:\1749\170\025\166\249\161\249\157\217\149\217\149\184\145\184\141\184\137\152\133\152\133\153\129\153\129\153\129\153\129\185\129\185\129\186\129\219\133\243\128\000\000\000\000\000\000\000\000\146\128\220\202\187\198\155\190z\186z\182Z\174:\170\025\166\025\162\249\153\217\149\217\145\185\141\185\137\184\133\152\133\152\129\153\129\153\129\153\129\153\129\185\129\185\129\186\129\186\129\219\129\244\128\000\000\000\000\000\000\000\000\146\128\220\198\187\194\155\190z\182Z\178:\170:\166\025\162\249\157\249\149\217\145\185\141\185\137\153\133\153\129\153\129\153\129\153\129\153\129\185\129\185\129\186\129\186\129\186\129\186\129\219\129\244\128\000\000\000\000\000\000\000\000\146\128\220\198\187\190\155\186z\178Z\174:\170\026\162\025\158\249\153\217\145\217\141\185\137\185\133\153\129\153\129\153\129\153\129\185\129\186\129\186\129\186\129\186\129\186\129\218\129\219\129\251\129\244\128\000\000\000\000\000\000\000\000\146\128\220\194\155\190{\182Z\178Z\170:\166\026\158\249\153\249\149\217\141\185\137\185\133\153\129\153\129\153\129\185\129\186\129\186\129\186\129\186\129\218\129\218\129\219\129\219\129\219\129\252\129\020\129\000\000\000\000\000\000\000\000\178\128\220\190\155\186{\178Z\174:\166:\162\026\158\250\149\217\145\217\137\185\133\185\129\185\129\185\129\186\129\186\129\186\129\186\129\218\129\218\129\219\129\219\129\219\129\219\129\251\129\028\130\021\129\000\000\000\000\000\000\000\000\179\128\188\190\155\186{\178[\170:\166:\158\026\154\250\145\218\141\218\137\185\129\185\129\186\129\186\129\186\129\186\129\218\129\219\129\219\129\219\129\219\129\219\129\251\129\252\129\252\129\029\130\021\129\000\000\000\000\000\000\000\000\179\128\188\190\155\182{\174[\170;\162:\158\026\150\250\145\218\137\218\133\186\129\186\129\186\129\186\129\218\129\219\129\219\129\219\129\219\129\251\129\251\129\252\129\252\129\252\129\252\129=\130\021\129\000\000\000\000\000\000\000\000\211\128\188\186\155\182{\174[\166;\162;\154\026\150\250\141\218\137\218\129\186\129\218\129\218\129\219\129\219\129\219\129\219\129\251\129\251\129\252\129\252\129\252\129\252\129\028\130\029\130>\130\021\129\000\000\000\000\000\000\000\000\211\128\188\186\155\178{\174[\166;\158;\154\027\146\251\141\218\133\218\129\218\129\219\129\219\129\219\129\219\129\251\129\252\129\252\129\252\129\252\129\252\129\028\130\029\130\029\130\029\130>\1305\129\000\000\000\000\000\000\000\000\212\128\189\186\156\178{\170[\166;\158;\154\027\146\251\137\251\133\219\129\219\129\219\129\219\129\251\129\252\129\252\129\252\129\252\129\028\130\028\130\029\130\029\130\029\130\029\130=\130^\1306\129\000\000\000\000\000\000\000\000\212\128\221\182\156\178|\170\092\166[\158;\150\027\146\251\137\251\133\219\129\251\129\251\129\252\129\252\129\252\129\252\129\028\130\028\130\029\130\029\130\029\130\029\130=\130>\130>\130_\1306\129\000\000\000\000\000\000\000\000\212\128\221\182\156\178|\170|\162\092\158<\150\028\146\027\138\251\133\251\129\252\129\252\129\252\129\252\129\028\130\029\130\029\130\029\130\029\130\029\130=\130>\130>\130>\130>\130\127\1306\129\000\000\000\000\000\000\000\000\244\128\221\182\188\178\156\170|\162\092\158<\150<\146\028\138\252\129\252\129\252\129\252\129\028\130\029\130\029\130\029\130\029\130=\130=\130>\130>\130>\130>\130^\130^\130\127\1306\129\000\000\000\000\000\000\000\000\244\128\221\182\189\178\156\170|\162|\158\092\150<\146<\138\028\134\028\130\028\130\029\130\029\130\029\130\029\130=\130=\130>\130>\130>\130>\130^\130_\130_\130_\130\127\130V\129\000\000\000\000\000\000\000\000\245\128\254\186\189\178\157\170\157\166}\158]\150]\146=\138\029\134\029\130\029\130\029\130\029\130=\130>\130>\130>\130>\130>\130^\130_\130_\130_\130_\130\127\130\159\130W\129\000\000\000\000\000\000\000\000\245\128\254\186\189\178\189\170\157\166}\158}\154]\146=\142=\134\029\130\029\130=\130>\130>\130>\130>\130^\130^\130_\130_\130_\130_\130\127\130\127\130\127\130\159\130W\129\000\000\000\000\000\000\000\000\021\129\254\186\221\182\189\174\157\166\157\162}\154}\150]\142=\138=\130>\130>\130>\130>\130^\130^\130_\130_\130_\130_\130\127\130\127\130\127\130\127\130\127\130\159\130W\129\000\000\000\000\000\000\000\000\021\129\031\187\222\182\222\174\190\170\158\162\158\158~\150~\146^\138^\134>\130>\130^\130_\130_\130_\130_\130_\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\159\130w\129\000\000\000\000\000\000\000\000\022\129\031\191\254\186\222\178\190\170\190\166\158\158\158\154~\146~\142^\134^\130_\130_\130_\130_\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\159\130x\129\000\000\000\000\000\000\000\0006\129\031\183\030\191\222\178\222\174\190\166\190\162\158\154\158\150\127\142\127\138\127\134_\130_\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\127\130\159\130x\129\000\000\000\000\000\000\000\0006\129\186\133\031\187?\191\031\183\031\175\255\170\223\162\223\154\191\150\191\146\191\138\159\134\159\130\159\130\159\130\159\130\159\130\159\130\159\130\159\130\159\130\159\130\159\130\159\130\127\130\219\129x\129\000\000\000\000\000\000\000\000T\1297\1296\1297\129W\129W\129W\129W\129W\129W\129W\129W\129W\129W\129w\129x\129w\129x\129x\129x\129x\129x\129x\129x\129x\129x\129x\129\153\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
-- #end

mediaStopSmall =
-- #begin
-- #include "res/mediaStop16.lua"

"\016\000\000\000\016\000\000\000\000\000\000\000 \000\000\000\016\000\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\178\128\020\1334\141\020\137\020\137\020\137\020\137\020\133\020\133\020\133\020\133\020\133\244\128\211\128\000\000\000\000\020\133}\231\252\210\187\198\154\186Z\1789\170\025\162\025\158\249\153\249\149\026\150;\154\020\129\000\000\000\0004\141\028\215\154\194Z\1829\170\025\162\249\153\217\145\185\141\185\137\185\133\185\133\218\137\021\129\000\000\000\0004\141\252\206z\186Z\174\025\162\249\153\217\141\185\137\153\129\153\129\153\129\186\129\219\1295\129\000\000\000\0004\137\220\198z\178:\166\249\153\217\145\185\133\153\129\185\129\186\129\186\129\218\129\251\1296\129\000\000\000\0005\137\188\190[\170\026\158\250\145\217\137\185\129\186\129\186\129\218\129\219\129\219\129\252\129V\129\000\000\000\000U\137\188\186[\166\026\154\250\141\186\129\186\129\219\129\219\129\219\129\252\129\252\129\029\130W\129\000\000\000\000U\137\188\186[\166\027\150\251\137\219\129\219\129\251\129\252\129\252\129\028\130\029\130>\130w\129\000\000\000\000V\137\221\182\092\162;\150\251\137\251\129\252\129\252\129\028\130\029\130\029\130=\130_\130x\129\000\000\000\000v\137\221\182|\162\092\150\028\134\252\129\029\130\029\130\029\130>\130>\130^\130\127\130x\129\000\000\000\000w\137\254\182\157\166]\150=\138\029\130=\130>\130>\130^\130_\130_\130\159\130\153\129\000\000\000\000\151\137\030\187\190\166~\154^\142>\130>\130^\130_\130_\130\127\130\127\130\159\130\153\129\000\000\000\000\152\133_\199\255\178\223\162\191\150\127\138\127\130\127\130\127\130\159\130\127\130\127\130\159\130\185\129\000\000\000\0006\129\152\137\185\137\185\133\185\133\153\129\153\129\185\129\153\129\186\129\186\129\186\129\186\129x\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"

-- #end

mediaArrowUpLarge =
"\016\000\000\000\016\000\000\000\000\000\000\000 \000\000\000\016\000\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\188\129\188\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\190\129}\166}\170\189\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\190\129\027\150\159\227?\207\251\145\157\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\157\129\218\137\159\223\254\190\222\186\222\190\186\137{\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\157\129\185\129\127\219\254\190\189\182\189\178\189\178\189\178\152\133\155\129\000\000\000\000\000\000\000\000\000\000<\129\153\129\127\219\031\203\189\178\157\178\157\174|\170|\170\157\170x\129:\129\000\000\000\000\000\000:\129\152\129\217\141\217\145;\158\221\182}\170|\166\092\166\149\145v\137w\133W\129z\129\000\000\000\000\000\000\000\000\000\000\000\000\217\149\030\199\092\162\092\162\092\162u\137\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\217\145\254\194<\158;\158;\154U\137\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\185\145\221\186;\154\027\150\027\150U\133\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\184\141\189\178\027\146\251\145\250\1414\133\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\152\141}\166\027\142\251\141\250\1374\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\021\1295\129\021\129\020\129\244\128\244\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
mediaArrowUpSmall =
"\016\000\000\000\016\000\000\000\000\000\000\000 \000\000\000\016\000\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\188\129\188\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\190\129}\166}\170\189\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\190\129\027\150\159\227?\207\251\145\157\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\157\129\218\137\159\223\254\190\222\186\222\190\186\137{\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\157\129\185\129\127\219\254\190\189\182\189\178\189\178\189\178\152\133\155\129\000\000\000\000\000\000\000\000\000\000<\129\153\129\127\219\031\203\189\178\157\178\157\174|\170|\170\157\170x\129:\129\000\000\000\000\000\000:\129\152\129\217\141\217\145;\158\221\182}\170|\166\092\166\149\145v\137w\133W\129z\129\000\000\000\000\000\000\000\000\000\000\000\000\217\149\030\199\092\162\092\162\092\162u\137\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\217\145\254\194<\158;\158;\154U\137\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\185\145\221\186;\154\027\150\027\150U\133\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\184\141\189\178\027\146\251\145\250\1414\133\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\152\141}\166\027\142\251\141\250\1374\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\021\1295\129\021\129\020\129\244\128\244\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"

mediaArrowDownLarge = 
"\016\000\000\000\016\000\000\000\000\000\000\000 \000\000\000\016\000\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\021\1295\129\021\129\021\129\021\129\244\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\184\145\254\190|\166<\154\250\1374\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\184\145\156\178\251\145\251\145\250\1414\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\217\149\189\186\027\154\027\150\026\150U\133\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\217\149\221\190<\158<\158;\154U\137\000\000\000\000\000\000\000\000\000\000\000\000\016\128[\129:\129\000\000\249\149\254\194\092\162\092\162;\158U\137\000\0006\129X\129\017\128\000\000\000\000z\129x\129\185\141\217\153[\170\030\199}\170|\166[\166\183\145v\137V\133W\129\245\128\000\000\000\000\000\000\156\129\153\133\031\199\031\199\222\190\157\178\157\174}\170\157\170\024\162W\129\027\129\000\000\000\000\000\000\000\000\000\000|\129\185\133\031\199\222\186\190\182\190\178\189\1788\170w\129\092\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\157\129\218\141?\203\254\190\222\186y\178\152\133\156\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\157\129\027\150_\207\154\190\184\141\156\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\189\129\248\157\249\153\157\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\188\129\188\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
mediaArrowDownSmall =
"\016\000\000\000\016\000\000\000\000\000\000\000 \000\000\000\016\000\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\021\1295\129\021\129\021\129\021\129\244\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\184\145\254\190|\166<\154\250\1374\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\184\145\156\178\251\145\251\145\250\1414\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\217\149\189\186\027\154\027\150\026\150U\133\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\217\149\221\190<\158<\158;\154U\137\000\000\000\000\000\000\000\000\000\000\000\000\016\128[\129:\129\000\000\249\149\254\194\092\162\092\162;\158U\137\000\0006\129X\129\017\128\000\000\000\000z\129x\129\185\141\217\153[\170\030\199}\170|\166[\166\183\145v\137V\133W\129\245\128\000\000\000\000\000\000\156\129\153\133\031\199\031\199\222\190\157\178\157\174}\170\157\170\024\162W\129\027\129\000\000\000\000\000\000\000\000\000\000|\129\185\133\031\199\222\186\190\182\190\178\189\1788\170w\129\092\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\157\129\218\141?\203\254\190\222\186y\178\152\133\156\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\157\129\027\150_\207\154\190\184\141\156\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\189\129\248\157\249\153\157\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\188\129\188\129\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"



function setComponentImagesSmall()
    
    mediaPlayImage=mediaPlaySmall
    mediaStopImage=mediaStopSmall
    mediaArrowUpImage = mediaArrowUpSmall
    mediaArrowDownImage = mediaArrowDownSmall
    
    componentSize = "small"
end

function setComponentImagesLarge()
    
    mediaPlayImage=mediaPlayLarge
    mediaStopImage=mediaStopLarge
    mediaArrowUpImage = mediaArrowUpLarge
    mediaArrowDownImage = mediaArrowDownLarge
    
    componentSize = "large"
end

function shouldUseSmallComponents()
    if platform.isDeviceModeRendering() or windowWidth < MIN_WIDTH_FOR_LARGE_COMPONENTS then
        return true
    else
        return false
    end
end

CONTROL_POINT_COLOR = {0,255,0}
DRAW_ORDER_CONTROL_POINT = 60
CONTROL_RADIUS = 3.5
ARROW_DIM = 3
BUTTON_SPACING_X = 2
BUTTON_SPACING_Y = 2
INPUT_CONTROL_HEIGHT = 17
INPUT_CONTROL_LABEL_HEIGHT = 15
OUTPUT_CONTROL_HEIGHT = 15
CONTROL_TEXT_HEIGHT = 10
CONTROL_INPUT_TEXT_HEIGHT = 15
CONTROL_TEXT_SIZE = 7
TEXT_SIZE_SCALE = 1
BUTTON_WIDTH=16
CONTROL_BAR_WIDTH = 80
CONTROL_BAR_WIDTH_SMALL = 80
CONTROL_BAR_WIDTH_LARGE = 160

if platform.hw()<7 then
    HANDHELD = true
    DELTA_MULTIPLIER = 5
else
    HANDHELD = false
    DELTA_MULTIPLIER = 1
end


function createControlPoint(xx,yy,onEnd,onMove)
   local cp =  components.circle()
                :__init(xx,yy, CONTROL_RADIUS)
                :setUseTransform(false)
       
                :setFilled(true)
                :setFillColor(unpack(CONTROL_POINT_COLOR))
                :setDrawOrder(DRAW_ORDER_CONTROL_POINT)
                :setDraggable{drag=true, onEnd=onEnd, onMove=onMove}
                :setOutlined(true)
                :setPenColor(0, 0, 0)
    return cp
end

function resetDragHandles(mat)
    if controlArray~=nil then
        for i=1,#controlArray do
    
            if controlArray[i].type == "dragHandle" then
                local x,y = controlArray[i].x, controlArray[i].y
                local p = transformCoord(mat,x,y)
                controlArray[i].handle:setPos(p.x,p.y)
            end
        end
    end
end

function setupComponents()



    if shouldUseSmallComponents() then                                    -- handheld or emulator
        setComponentImagesSmall()
        TEXT_SIZE_SCALE = 1
        BUTTON_WIDTH = 16
    else                                                                        -- desktop
        setComponentImagesLarge()
        TEXT_SIZE_SCALE = 2
        BUTTON_WIDTH = 32
    end
    
    local cbw = CONTROL_BAR_WIDTH*TEXT_SIZE_SCALE
    local cbx0 = windowWidth - cbw
    local bsx = BUTTON_SPACING_X*TEXT_SIZE_SCALE
    
    objStatusBar = components.rect()
        :__init(cbx0, 0, cbw, windowHeight)
        :setUseTransform(false)
        :setFilled(true)
        :setFillColor(224,224,224)
        :setOutlined(true)
        :setPenColor(0, 0, 0)
        :setPenStyle("thin", "smooth")
        :setDrawOrder(DRAW_ORDER_STATUS)
	
    local focusOrder=0
    
   objScreen = components.rect()
        :__init(0, 0, cbx0, windowHeight)
        :setUseTransform(false)
        :setFilled(false)
        :setOutlined(true)
        :setPenColor(0, 0, 0)
        :setPenStyle("thin", "smooth")
        :setDrawOrder(DRAW_ORDER_SCREEN_RECT)
	:setFocusOrder(focusOrder)
	
	objScreen.onCharIn = onCharInScreen
	
	focusOrder = focusOrder+1    

   local controlY = BUTTON_SPACING_Y
    for i=1,#controlArray do
		controlArray[i].y0 = controlY
		if controlArray[i].type == "textInput" then
			controlArray[i].objLabel = components.text()
			:__init(cbx0 +bsx,controlY, cbw - 2*bsx, CONTROL_TEXT_HEIGHT*TEXT_SIZE_SCALE, controlArray[i].label, "left", "middle")
			:setUseTransform(false)
			:setTextStyle("sansserif", "r", CONTROL_TEXT_SIZE*TEXT_SIZE_SCALE)
			:setDrawOrder(DRAW_ORDER_CONTROL)
			controlY = controlY + INPUT_CONTROL_LABEL_HEIGHT*TEXT_SIZE_SCALE

			controlArray[i].objValue = components.textEdit()
			:__init(cbx0 +bsx,controlY, cbw - 2*bsx, CONTROL_INPUT_TEXT_HEIGHT*TEXT_SIZE_SCALE, string.format("%g",controlArray[i].value), "left", "top")
			:setUseTransform(false)
			:setTextStyle("sansserif", "r", CONTROL_TEXT_SIZE*TEXT_SIZE_SCALE)
			:setDrawOrder(DRAW_ORDER_CONTROL)
			:setOutlined(true)
			:setFocusOrder(focusOrder)
			:setSelectable{select=true, onClick=onClickObject}
			:setFilled(true)
			:setFillColor(255,255,255)
					controlArray[i].objValue.onEnterKey = onAcceptValue
				focusOrder=focusOrder+1
			controlY = controlY + INPUT_CONTROL_HEIGHT*TEXT_SIZE_SCALE
		elseif controlArray[i].type == "functionTextInput" then
			controlArray[i].objLabel = components.text()
			:__init(cbx0 +bsx,controlY, cbw - 2*bsx, CONTROL_TEXT_HEIGHT*TEXT_SIZE_SCALE, controlArray[i].label, "left", "middle")
			:setUseTransform(false)
			:setTextStyle("sansserif", "r", CONTROL_TEXT_SIZE*TEXT_SIZE_SCALE)
			:setDrawOrder(DRAW_ORDER_CONTROL)
			controlY = controlY + INPUT_CONTROL_LABEL_HEIGHT*TEXT_SIZE_SCALE

			controlArray[i].objValue = components.textEdit()
			:__init(cbx0 +bsx,controlY, cbw - 2*bsx, CONTROL_INPUT_TEXT_HEIGHT*TEXT_SIZE_SCALE, string.format("%s",controlArray[i].value), "left", "top")
			:setUseTransform(false)
			:setTextStyle("sansserif", "r", CONTROL_TEXT_SIZE*TEXT_SIZE_SCALE)
			:setDrawOrder(DRAW_ORDER_CONTROL)
			:setOutlined(true)
			:setFocusOrder(focusOrder)
			:setSelectable{select=true, onClick=onClickObject}
			:setFilled(true)
			:setFillColor(255,255,255)
					controlArray[i].objValue.onEnterKey = onAcceptValue
					controlArray[i].objValue.maxChars = 200

				focusOrder=focusOrder+1
			controlY = controlY + INPUT_CONTROL_HEIGHT*TEXT_SIZE_SCALE
		    
		elseif controlArray[i].type == "slider" then
			controlArray[i].objLabel = components.text()
			:__init(cbx0+bsx,controlY, cbw-2*bsx, CONTROL_TEXT_HEIGHT*TEXT_SIZE_SCALE, controlArray[i].label, "left", "middle")
			:setUseTransform(false)
			:setTextStyle("sansserif", "r", CONTROL_TEXT_SIZE*TEXT_SIZE_SCALE)
			:setDrawOrder(DRAW_ORDER_CONTROL)
			controlY = controlY + INPUT_CONTROL_LABEL_HEIGHT*TEXT_SIZE_SCALE

			controlArray[i].objValue = components.slider()
			:__init(cbx0+bsx,controlY, cbw - 4*bsx-BUTTON_WIDTH, CONTROL_INPUT_TEXT_HEIGHT*TEXT_SIZE_SCALE)
			:setUseTransform(false)
			:setTextStyle("sansserif", "r", CONTROL_TEXT_SIZE*TEXT_SIZE_SCALE)
			:setDrawOrder(DRAW_ORDER_CONTROL)
			:setFillColor(128, 128, 255)
			:setPenColor(0, 0, 0)
			:setPenStyle("thin", "smooth")
			:setOutlined(false)
			:setFocusOrder(focusOrder)
				:setRange(controlArray[i].min, controlArray[i].max)
				:setDivisions(4)
			:setValue(controlArray[i].value)
			
			:setSelectable{select=true, onClick=onClickObject}
			controlArray[i].objValue.onEnterKey = onAcceptValue
				focusOrder=focusOrder+1
			
		controlArray[i].objPlayButton = components.rect()
			:__init(windowWidth-BUTTON_WIDTH-bsx, controlY, BUTTON_WIDTH, BUTTON_WIDTH)
			:setImage(image.new(mediaPlayImage))
			:setUseTransform(false)
			:setFocusOrder(focusOrder)
			:setData("state", "play")
			:setData("control",controlArray[i])
			:setSelectable{select=true, onClick=onClickPlayButton}
			:setDrawOrder(DRAW_ORDER_CONTROL)
			controlArray[i].objPlayButton.onEnterKey = function(foo) onClickPlayButton(foo) end	    
			focusOrder=focusOrder+1
	       
			
			controlY = controlY + INPUT_CONTROL_HEIGHT*TEXT_SIZE_SCALE

		elseif controlArray[i].type == "output" then
			controlArray[i].objLabel = components.text()
			:__init(cbx0+bsx,controlY, cbw - 2*bsx, CONTROL_TEXT_HEIGHT*TEXT_SIZE_SCALE, controlArray[i].label, "left", "middle")
			:setUseTransform(false)
			:setTextStyle("sansserif", "r", CONTROL_TEXT_SIZE*TEXT_SIZE_SCALE)
			:setDrawOrder(DRAW_ORDER_CONTROL)
			controlY = controlY + INPUT_CONTROL_LABEL_HEIGHT*TEXT_SIZE_SCALE

			controlArray[i].objValue = components.text()
			:__init(cbx0+bsx,controlY, cbw - 2*bsx, CONTROL_TEXT_HEIGHT*TEXT_SIZE_SCALE, string.format("%g",controlArray[i].value), "left", "middle")
			:setUseTransform(false)
			:setTextStyle("sansserif", "r", CONTROL_TEXT_SIZE*TEXT_SIZE_SCALE)
			:setDrawOrder(DRAW_ORDER_CONTROL)
			controlY = controlY + OUTPUT_CONTROL_HEIGHT*TEXT_SIZE_SCALE
	            
			elseif controlArray[i].type == "dragHandle" then
				controlArray[i].handle = createControlPoint(controlArray[i].x,controlArray[i].y,controlArray[i].onEnd,controlArray[i].onMove)
		end
		controlArray[i].y1 = controlY
    end

    focusOrder=focusOrder+1
    haveSetupComponents = true
end



function loadComponentImages()

end

function recalcControlPositions()
	local depth = {}
	depth["textInput"] = INPUT_CONTROL_LABEL_HEIGHT*TEXT_SIZE_SCALE + INPUT_CONTROL_HEIGHT*TEXT_SIZE_SCALE 
	depth["functionTextInput"] = INPUT_CONTROL_LABEL_HEIGHT*TEXT_SIZE_SCALE + INPUT_CONTROL_HEIGHT*TEXT_SIZE_SCALE 
	depth["slider"] = INPUT_CONTROL_LABEL_HEIGHT*TEXT_SIZE_SCALE + INPUT_CONTROL_HEIGHT*TEXT_SIZE_SCALE 
	depth["output"] = INPUT_CONTROL_LABEL_HEIGHT*TEXT_SIZE_SCALE + OUTPUT_CONTROL_HEIGHT*TEXT_SIZE_SCALE
        depth["dragHandle"] = 0
	
	local yy = 0
	for _,control in ipairs(controlArray) do
                
		control.y0 = yy
		yy = yy +depth[control.type]
		control.y1 = yy
	end
end

function updateComponents()
    if shouldUseSmallComponents() then                                    -- handheld or emulator
        setComponentImagesSmall()
        TEXT_SIZE_SCALE = 1
        BUTTON_WIDTH = 16
    else                                                                        -- desktop
        setComponentImagesLarge()
        TEXT_SIZE_SCALE = 2
        BUTTON_WIDTH = 32
    end
    
    local cbw = CONTROL_BAR_WIDTH*TEXT_SIZE_SCALE
    local cbx0 = windowWidth - cbw
    local bsx = BUTTON_SPACING_X*TEXT_SIZE_SCALE
        
    if objStatusBar ~= nil then
        objStatusBar:setSize(cbw, windowHeight)
        objStatusBar:setPos(windowWidth - cbw , 0)
   end
    if objScreen ~= nil then
        objScreen:setSize(cbx0, windowHeight)
        objScreen:setPos(0 , 0)
   end
   
   recalcControlPositions()  -- may have changed due to jump from small to large text size
    
    local controlHeight=0
    if #controlArray > 0 then
        controlHeight =  controlArray[#controlArray].y1 - controlArray[1].y0
    end
    if controlHeight > windowHeight - 2*BUTTON_SPACING_Y then
		shouldScroll = true
	else
		shouldScroll = false
	end 
    local focusOrder = 1
    local controlY = BUTTON_SPACING_Y
    local endY = windowHeight - BUTTON_SPACING_Y
    if controlArray.visibleStart == nil then controlArray.visibleStart = 1 end
	if shouldScroll then
		if controlArray.objUpArrow ~= nil then
			controlArray.objUpArrow:setPos(cbx0 +(cbw-BUTTON_WIDTH)/2, BUTTON_SPACING_Y)
			controlArray.objDownArrow:setPos(cbx0 +(cbw-BUTTON_WIDTH)/2, windowHeight - BUTTON_SPACING_Y-BUTTON_WIDTH)
			controlArray.objUpArrow:setFocusOrder(focusOrder)
			focusOrder = focusOrder+1
		else
			controlArray.objUpArrow = components.rect()
				:__init(cbx0 +(cbw-BUTTON_WIDTH)/2, BUTTON_SPACING_Y, BUTTON_WIDTH, BUTTON_WIDTH)
				:setImage(image.new(mediaArrowUpImage))
				:setUseTransform(false)
				:setFocusOrder(focusOrder)
				:setSelectable{select=true, onClick=onClickObject}
				:setDrawOrder(DRAW_ORDER_CONTROL)
				controlArray.objUpArrow.onEnterKey = function(foo) onClickControlUp(foo) end	    
				focusOrder=focusOrder+1
			controlArray.objDownArrow = components.rect()
				:__init(cbx0 +(cbw-BUTTON_WIDTH)/2, windowHeight - BUTTON_SPACING_Y-BUTTON_WIDTH, BUTTON_WIDTH, BUTTON_WIDTH)
				:setImage(image.new(mediaArrowDownImage))
				:setUseTransform(false)
				:setSelectable{select=true, onClick=onClickObject}
				:setDrawOrder(DRAW_ORDER_CONTROL)
				controlArray.objDownArrow.onEnterKey = function(foo) onClickControlDown(foo) end	    
		end
		controlY = controlY +BUTTON_WIDTH + BUTTON_SPACING_Y
		endY = endY -2*BUTTON_WIDTH - 2*BUTTON_SPACING_Y
	else
		if controlArray.objUpArrow ~= nil then
			components:remove(controlArray.objUpArrow)
			controlArray.objUpArrow	= nil
		end
		if controlArray.objDownArrow ~= nil then
			components:remove(controlArray.objDownArrow)
			controlArray.objDownArrow	= nil
		end
	end    

    for i,control in ipairs(controlArray) do

		local isVisible = true
		if i < controlArray.visibleStart then
			 isVisible = false
		elseif controlArray[i].y1 - controlArray[controlArray.visibleStart].y0 > endY then
			isVisible = false
		end
		if control.objLabel then control.objLabel.visible = isVisible end
		if control.objValue then control.objValue.visible = isVisible end
		
		if isVisible then
			if control.type == "textInput" or control.type == "functionTextInput" then
				control.objLabel:setSize(cbw - 2*bsx, CONTROL_TEXT_HEIGHT*TEXT_SIZE_SCALE)
				control.objLabel:setPos(cbx0 +bsx,controlY)
				control.objLabel:setTextStyle("sansserif", "r", CONTROL_TEXT_SIZE*TEXT_SIZE_SCALE)
				controlY = controlY + INPUT_CONTROL_LABEL_HEIGHT*TEXT_SIZE_SCALE
				control.objValue:setSize(cbw - 2*bsx, CONTROL_INPUT_TEXT_HEIGHT*TEXT_SIZE_SCALE)
				control.objValue:setPos(cbx0 +bsx,controlY)
				control.objValue:setTextStyle("sansserif", "r", CONTROL_TEXT_SIZE*TEXT_SIZE_SCALE)			
				control.objValue:setFocusOrder(focusOrder)
				control.objValue:setSelectable{select=true, onClick=onClickObject}
				focusOrder = focusOrder +1

				controlY = controlY + INPUT_CONTROL_HEIGHT*TEXT_SIZE_SCALE
			    
			elseif control.type == "slider" then
				control.objLabel:setSize(cbw - 2*bsx, CONTROL_TEXT_HEIGHT*TEXT_SIZE_SCALE)
				control.objLabel:setPos(cbx0 +bsx,controlY)
				control.objLabel:setTextStyle("sansserif", "r", CONTROL_TEXT_SIZE*TEXT_SIZE_SCALE)
				controlY = controlY + INPUT_CONTROL_LABEL_HEIGHT*TEXT_SIZE_SCALE
				control.objValue:setSize(cbw - 4*bsx-BUTTON_WIDTH, CONTROL_INPUT_TEXT_HEIGHT*TEXT_SIZE_SCALE)
				control.objValue:setPos(cbx0 +bsx,controlY)
				control.objValue:setTextStyle("sansserif", "r", CONTROL_TEXT_SIZE*TEXT_SIZE_SCALE)
				control.objValue:setFocusOrder(focusOrder)
				control.objValue:setSelectable{select=true, onClick=onClickObject}
				focusOrder = focusOrder +1
				
				control.objPlayButton:setSize(BUTTON_WIDTH, BUTTON_WIDTH)
				control.objPlayButton:setPos(windowWidth-BUTTON_WIDTH-bsx, controlY)
					if control.objPlayButton:getData("state") == "play" then
						control.objPlayButton:setImage(image.new(mediaPlayImage))
					else
						control.objPlayButton:setImage(image.new(mediaStopImage))
					end
				control.objPlayButton:setFocusOrder(focusOrder)
				focusOrder = focusOrder +1
				
				controlY = controlY + INPUT_CONTROL_HEIGHT*TEXT_SIZE_SCALE
				control.objPlayButton.visible = isVisible

			elseif control.type == "output" then
		        
 				control.objLabel:setSize(cbw - 2*bsx, CONTROL_TEXT_HEIGHT*TEXT_SIZE_SCALE)
				control.objLabel:setPos(cbx0 +bsx,controlY)
				control.objLabel:setTextStyle("sansserif", "r", CONTROL_TEXT_SIZE*TEXT_SIZE_SCALE)
				controlY = controlY + INPUT_CONTROL_LABEL_HEIGHT*TEXT_SIZE_SCALE
				control.objValue:setSize(cbw - 2*bsx, CONTROL_TEXT_HEIGHT*TEXT_SIZE_SCALE)
				control.objValue:setPos(cbx0 +bsx,controlY)
				control.objValue:setTextStyle("sansserif", "r", CONTROL_TEXT_SIZE*TEXT_SIZE_SCALE)
		       
				controlY = controlY + OUTPUT_CONTROL_HEIGHT*TEXT_SIZE_SCALE  
			elseif control.type == "dragHandle" then
				control.handle:setSize(CONTROL_RADIUS)
			end
		else
			if control.type == "slider" then	
				control.objPlayButton.visible = isVisible		
				control.objValue:setFocusOrder(-1)		
				control.objPlayButton:setFocusOrder(-1)		
				control.objValue:setSelectable{select=false}
				control.objPlayButton:setSelectable{select=false}
			elseif control.type == "textInput" or control.type == "functionTextInput" then
				control.objValue:setFocusOrder(-1)		
				control.objValue:setSelectable{select=false}
			end
		end
    end
	if shouldScroll and controlArray.objDownArrow ~=nil then
		controlArray.objDownArrow:setFocusOrder(focusOrder)
		focusOrder = focusOrder + 1
	end    
	components:advanceFocus(0)     -- reposition focus
end

function calculateDimensions()
    windowWidth=platform.window:width()
    windowHeight=platform.window:height()
    
    if windowWidth==0 then windowWidth=318 end
    if windowHeight==0 then windowHeight=212 end
        
        
    if shouldUseSmallComponents() then                                    -- handheld or emulator
    else                                                                        -- desktop
     end
    
end

-- *********************
-- **** entry point ****
-- *********************
function initApp()
 
    calculateDimensions()    

	addControls()

   
    setupComponents()                                                           -- creates all the static components
    updateComponents()                                                          -- update the layout
    components:advanceFocus(0)                                                  -- init focus to the first item
    components:updateDrawOrder()                                                -- sort objects into draw order
    setCommonGlobal(windowWidth,windowHeight)

end




function setCommonGlobal(w,h)
    if commonGlobal ~= nil then
        commonGlobal.canvas.width = w - CONTROL_BAR_WIDTH*TEXT_SIZE_SCALE  
        commonGlobal.canvas.height = h
	commonGlobal.screenBox = BoundingBox();
	commonGlobal.screenBox:extend(0,0);
	commonGlobal.screenBox:extend(commonGlobal.canvas.width, commonGlobal.canvas.height);
	commonGlobal.world = BoundingBox();
    end

end


function onResize(w, h)

	if app_initialized == nil then
		setCommonGlobal(w,h)
		app_initialized = true
		initApp()
        else
            calculateDimensions()    

            updateComponents()
		setCommonGlobal(w,h)

	end
        
        

    if commonGlobal ~= nil then
	doFullUpdate(commonGlobal.world, commonGlobal.screenBox, commonGlobal.canvas, commonGlobal.context, commonGlobal.world2Screen, commonGlobal.screen2World);
    end
end





on.addListener("paint", onPaint, on.PRIORITY_MAX)
on.addListener("resize", onResize)


load()

